/*
 * abmail.c - abuse mail filter
 *
 * Copyright (C) 2006-2008 Taiji Yamada <taiji@aihara.co.jp>
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Affero General Public License for more details.
 */
/*
 * usage:
 *	$ ./abmail [-t] [-b|-a] [-g [-r]] [-d] [-v|-v-|-vv] [-vvv] [-vvvv] [-c] [-D] [--sm-client[=sm_pathname]] [-M-] [-MM[-]] [-P-] < mail; test $? = 1 && echo "abmail"
 *	$ ./abmail [..] [+|- mail|mhdir|maildir [...]]
 *	$ ./abmail [..] --sm-server[=sm_pathname]|--sm-delete[=sm_pathname]
 *	$ ./abmail [..] --am-daemon=server_host|--am-dbfile=path_to.bdb
 */
#include <stdio.h>	/* stderr, etc */
#include <stdlib.h>	/* qsort, bsearch, etc */
#include <stdarg.h>	/* va_list, etc */
#include <string.h>	/* strcmp, etc */
#include <ctype.h>	/* isspace, etc */
#include <libgen.h>	/* dirname, etc */
#include <sys/types.h>	/* netinet/in.h, arpa/inet.h, sys/socket.h, etc */
#include <sys/stat.h>	/* stat, etc */
#include <dirent.h>	/* opendir, etc */
#include <netinet/in.h>	/* in_addr, etc */
#include <arpa/inet.h>	/* inet_ntoa, etc */
#include <pcre.h>	/* pcre, etc */
#include <netdb.h>	/* gethostbyname, etc */
#if !defined(DISABLE_MMAN)
#include <sys/mman.h>	/* mmap, etc */
#include <fcntl.h>	/* O_CREATE, etc */
#include <unistd.h>	/* ftruncate, etc */
#endif
#if !defined(DISABLE_PTHREAD)
#include <pthread.h>	/* pthread_create, etc */
#endif
#include <sys/param.h>	/* MAXPATHLEN, etc */
#include <sys/socket.h>	/* AN_INET, etc */
#if defined(USE_DRACD_DB)
#include <sys/fcntl.h>	/* O_RDONLY, etc */
#define DB185 DB
#include <db_185.h>	/* dbopen, etc */
#endif
#include "abcidr.h"
#include "abmisc.h"
#if defined(USE_ABADDR_DB)
#include "abaddri.h"
#endif

#define min(a, b) (((a)<(b))?(a):(b))
#define max(a, b) (((a)>(b))?(a):(b))
#define lbu(l, b, u) min(max((l), (b)), (u))

#if !defined(HAVE_STRLCPY)
#define strlcpy(d, s, z) do { strncpy(d, s, z); if ((z) > 0) (d)[(z)-1] = '\0'; } while (0)
#endif

#if !defined(HAVE_STRLCAT)
#define strlcat(d, s, z) do { if ((z) > strlen(d)) strncat(d, s, (z)-strlen(d)-1); } while (0)
#endif

#if !defined(HAVE_INET_ATON)
#define inet_aton(s, p) ((((p)->s_addr=inet_addr(s))!=-1||strcmp(s, "255.255.255.255")==0)?1:0)
#endif

#if !defined(PREFIX)
#define PREFIX "/opt/local"
#endif

int aredigit(const char *s)
{
  size_t i, z;

  if (!s) return 0;
  if (!(z = strlen(s)))
    return 0;
  for (i=0; i<z; i++)
    if (!isdigit(s[i]))
      return 0;
  return !0;
}

char *lookup_reverse(const char *a, char *b, size_t z, int paranoid)
{
  char *rv = NULL, s[2048];
#if !defined(DISABLE_PTHREAD)
  struct addrinfo hints, *res, *r;
  int err, flag = 0;

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags |= AI_NUMERICHOST;
  if (!(err=getaddrinfo(a, NULL, &hints, &res))) {
    for (r=res; r!=0; r=r->ai_next) {
      if (!(err=getnameinfo(r->ai_addr, r->ai_addrlen, s, sizeof(s),
			    NULL, 0, NI_NAMEREQD))) {
	if (!paranoid) {
	  strlcpy(b, s, z);
	  rv = b;
	  break;
	}
	else {
	  struct addrinfo hints, *res, *q;
	  int err;

	  memset(&hints, 0, sizeof(hints));
	  hints.ai_family = AF_UNSPEC;
	  hints.ai_socktype = SOCK_STREAM;
	  if (!(err=getaddrinfo(s, NULL, &hints, &res))) {
	    for (q=res; q!=0; q=q->ai_next) {
	      if (r->ai_family == AF_INET && r->ai_family == q->ai_family &&
		  ((struct sockaddr_in *)r->ai_addr)->sin_addr.s_addr ==
		  ((struct sockaddr_in *)q->ai_addr)->sin_addr.s_addr) {
		flag = !0;
		break;
	      }
#if !defined(DISABLE_IPv6)
	      if (r->ai_family == AF_INET6 && r->ai_family == q->ai_family &&
		  memcmp(&((struct sockaddr_in6 *)r->ai_addr)->sin6_addr,
			 &((struct sockaddr_in6 *)q->ai_addr)->sin6_addr,
			 sizeof(struct in6_addr)) == 0) {
		flag = !0;
		break;
	      }
#endif
	    }
	    freeaddrinfo(res);
	  }
	  if (flag) {
	    strlcpy(b, s, z);
	    rv = b;
	    break;
	  }
	}
      }
    }
    freeaddrinfo(res);
  }
#else
  struct hostent *he;
  struct in_addr in;
#if !defined(DISABLE_IPv6)
  struct in6_addr in6;
#endif

  in.s_addr = -1;
  if ((inet_aton(a, &in) != 0 &&
       (he=gethostbyaddr((char *)&in, sizeof(in), AF_INET)))
#if !defined(DISABLE_IPv6)
      ||
      (inet_pton(AF_INET6, a, &in6) != 0 &&
       (he=gethostbyaddr((char *)&in6, sizeof(in6), AF_INET6)))
#endif
      ) {
    strlcpy(s, he->h_name, sizeof(s));
    if (!paranoid) {
      strlcpy(b, s, z);
      rv = b;
    }
    else {
      struct hostent *he;
      int i, flag = 0;

      if (in.s_addr != -1 && (he=gethostbyname(s)))
	for (i=0; he->h_addr_list[i]; i++)
	  if (in.s_addr == (*(struct in_addr *)he->h_addr_list[i]).s_addr) {
	    flag = !0;
	    break;
	  }
#if !defined(DISABLE_IPv6)
      if (!flag && (he=gethostbyname2(s, AF_INET6)))
	for (i=0; he->h_addr_list[i]; i++)
	  if (memcmp(in6.s6_addr,
		     ((struct in6_addr *)he->h_addr_list[i])->s6_addr,
		     sizeof(struct in6_addr)) == 0) {
	    flag = !0;
	    break;
	  }
#endif
      if (flag) {
	strlcpy(b, s, z);
	rv = b;
      }
    }
  }
#endif
  return rv;
}

#define echo_msg(msg) fprintf(m->msgout, "abmail: %s\n", msg)
#define this_is_abmail(msg) do { rc = 1; if (m->verbose&(1<<0)) echo_msg(msg); } while (0)
#define this_is_not_abmail(msg) do { rc = -1; if (m->verbose&(1<<1)) echo_msg(msg); } while (0)
#define this_was_not_abmail(msg) do { if (rc == 0) if (m->verbose&(1<<1)) echo_msg(msg); } while (0)

#define echo_msg_helo(msg) fprintf(m->msgout, "abmail: %s; %s([%s]%s)%s\n", msg, m->loh?"bounced":"", address, name, fqdn)
#define this_is_abhelo(msg) do { rc = 1; if (m->verbose&(1<<0)) { if (m->verbose&(1<<2)&&n>3) echo_msg_helo(msg); else echo_msg(msg); } } while (0)
#define this_is_not_abhelo(msg) do { rc = -1; if (m->verbose&(1<<1)) { if (m->verbose&(1<<2)&&n>3) echo_msg_helo(msg); else echo_msg(msg); } } while (0)

#define echo_abmail_helo() fprintf(m->msgout, "abmail-helo: %s([%s]%s)%s\n", m->loh?"bounced":"", address, name, fqdn);
#define echo_abmail_dnsbl() fprintf(m->msgout, "abmail-dnsbl: %s\n", dnsbls[i]);

#if defined(USE_DRACD_DB)
#if !defined(DRACD_DB)
#define DRACD_DB "/etc/postfix/dracd.db"
#endif
int dracd_find(const char *ip)
{
  int rv = 0;
  DB *dbp;
  DBT dbk, dbd;

  if (!(dbp=dbopen(DRACD_DB, O_RDONLY, 0644, DB_HASH, NULL))) {
    fprintf(stderr, "dbopen failed\n");
    return rv;
  }
  dbk.data = (void *)ip;
  dbk.size = strlen(ip);
  if (dbp->get(dbp, &dbk, &dbd, 0) == RET_SUCCESS)
    rv = 1;
  dbp->close(dbp);
  return rv;
}
#endif

const char *dnsbls[] = {
  NULL,
  /*"list.dsbl.org",*/		/* expired */
  /*"unconfirmed.dsbl.org",*/	/* expired */
  /*"spam.dnsbl.sorbs.net",*/	/* too unsparing */
  "cbl.abuseat.org",
  /*"opm.blitzed.org",*/	/* expired */
  /*"dnsbl.sorbs.net",*/	/* too unsparing */
  "bl.spamcop.net",
  /*"spam.tqmcube.com",*/	/* expired, unsparing */
  /*"bl.spamcannibal.org",*/	/* too unsparing */
  "dnsbl-1.uceprotect.net",
  //"dnsbl-2.uceprotect.net",
  //"dnsbl-3.uceprotect.net",
  "zen.spamhaus.org",
  //"niku.2ch.net",
};

#if !defined(DISABLE_PTHREAD)
struct thr_dnsbl_find_arg {
  char name[2048];
  int i, *rv, *n;
  pthread_t *selves;
  pthread_mutex_t *mutex;
};

int thr_dnsbl_find_enable = !0;

void *thr_dnsbl_find(void *arg)
{
  int rv = 0;
  struct thr_dnsbl_find_arg *a = (struct thr_dnsbl_find_arg *)arg;
  struct addrinfo hints, *res;

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = AF_INET;
  hints.ai_socktype = SOCK_STREAM;
  if ((rv = (getaddrinfo(a->name, NULL, &hints, &res) == 0)?a->i:0)) {
    freeaddrinfo(res);
    if (*a->rv == 0) {
      pthread_mutex_lock(a->mutex);
      *a->rv = rv;
      pthread_mutex_unlock(a->mutex);
    }
  }
  return NULL;
}
#endif

int dnsbl_find(const char *ip)
{
  int rv = 0;
  int i;
  unsigned int ip4[4];

#if !defined(DISABLE_PTHREAD)
  if (!thr_dnsbl_find_enable) {
#endif
    struct hostent *he;
    char name[2048];

    if (sscanf(ip, "%u.%u.%u.%u", &ip4[0], &ip4[1], &ip4[2], &ip4[3]) != 4)
      return rv;
    for (i=1; i<sizeof(dnsbls)/sizeof(char *); i++) {
      sprintf(name, "%u.%u.%u.%u.%s",
	      ip4[3], ip4[2], ip4[1], ip4[0], dnsbls[i]);	/* not insecure */
      if ((he=gethostbyname(name))) {
	rv = i;
	break;
      }
    }
#if !defined(DISABLE_PTHREAD)
  }
  else {
    int n = 1;
    pthread_t threads[sizeof(dnsbls)/sizeof(char *)];
    pthread_mutex_t mutex;
    struct thr_dnsbl_find_arg args[sizeof(dnsbls)/sizeof(char *)];

    if (sscanf(ip, "%u.%u.%u.%u", &ip4[0], &ip4[1], &ip4[2], &ip4[3]) != 4)
      return rv;
    pthread_mutex_init(&mutex, NULL);
    for (i=1; i<sizeof(dnsbls)/sizeof(char *) && rv == 0; i++) {
      snprintf(args[i].name, sizeof(args[i].name), "%u.%u.%u.%u.%s",
	       ip4[3], ip4[2], ip4[1], ip4[0], dnsbls[i]);
      args[i].i = i;
      args[i].rv = &rv;
      args[i].n = &n;
      args[i].selves = threads;
      args[i].mutex = &mutex;
      pthread_create(&threads[i], NULL, thr_dnsbl_find, (void *)&args[i]);
      n++;
    }
    if (rv) {
      pthread_join(threads[rv], NULL);
      for (i=1; i<n; i++)
	if (i != rv) pthread_detach(threads[i]);
    }
    else
      for (i=1; i<n; i++)
	pthread_join(threads[i], NULL);
  }
#endif
  return rv;
}

struct rs {
  char *se;
  int lb, ub;
};

static
struct abmail {
  FILE *msgout;
  int verbose, paranoid, dnsbl;
#if defined(USE_ABADDR_DB)
  int abaddr;
  char *host, *file;
#endif
  int body;		/* process body */
  int lom, loh, hop;	/* level of multipart, level of header, header of part */
  char from[3][2048];	/* 0: full address, 1: logname, 2: domain */
  char boundary[1][2048];
  char buf[32768], line[32768];
  size_t nrs; struct rs *rss;
  struct amset *as;
#if !defined(DISABLE_IPv6)
  struct am6set *a6s;
#endif
} me = {
  NULL/*stderr*/,
  0, !0, 0,
#if defined(USE_ABADDR_DB)
  !0,
  NULL, NULL,
#endif
  0,
  0, 0, 0,
  { "", "", "", }, { "" },
  "", "",
  0, NULL,
  &amset,
#if !defined(DISABLE_IPv6)
  &am6set,
#endif
};

struct pz {
  char *p;
  size_t z;
};

int pcre_set_substring_pz_array(char *subject, int *ovector,
				int stringcount, struct pz a[])
{
  int i;

  for (i=0; i<stringcount; i++) {
    a[i].p = subject + ovector[i*2];
    a[i].z = ovector[i*2+1] - ovector[i*2];
  }
  return 0;
}

int abadr_find(struct abmail *m, const char *a, int *rc)
{
  struct in_addr in;
  struct am *am = NULL;
#if !defined(DISABLE_IPv6)
  struct in6_addr in6;
  struct am6 *am6 = NULL;
  int af = AF_INET;
#define case_af(a6,a) ((af==AF_INET6)?(a6):(a))
#else
#define case_af(a6,a) (a)
#if defined(USE_ABADDR_DB)
  int af = AF_INET;
#endif
#endif
#define am_rc case_af(am6->rc, am->rc)

  if (inet_aton(a, &in) == 0)
#if !defined(DISABLE_IPv6)
    {
      if (inet_pton(AF_INET6, a, &in6) == 0)
	return -1;
      else
	af = AF_INET6;
    }
#else
    return -1;
#endif
#if defined(USE_DRACD_DB)
  if (dracd_find(a)) {
    if (rc) *rc = -1;
    return 0;
  }
#endif
#if defined(USE_ABADDR_DB)
  if (m->abaddr) {
    char *a = case_af((char *)&in6, (char *)&in);
    int f = AF_UNSPEC, c = 0;

    if (abaddr_addr_get(m->host, m->file, af, a, a, &f, &c, NULL) == 0) {
      if (f != AF_UNSPEC) {
	if (rc) *rc = c;
	return 0;
      }
      return 1;
    }
  }
#endif
  if (case_af((void *)(am6=am6_search(m->a6s->nam, m->a6s->ams, in6)),
	      (void *)(am=am_search(m->as->nam, m->as->ams, in)))) {
    if (rc) *rc = am_rc;
    return 0;
  }
  return 1;
}

int amp_to(struct abmail *m, int n, const char **l, struct pz a[]);
int amp_from(struct abmail *m, int n, const char **l, struct pz a[]);
int amp_bofrom(struct abmail *m, int n, const char **l, struct pz a[]);
int amp_local(struct abmail *m, int n, const char **l, struct pz a[]);
int amp_auth(struct abmail *m, int n, const char **l, struct pz a[]);
int amp_subst(struct abmail *m, int n, const char **l, struct pz a[]);
int amp_abhelo(struct abmail *m, int n, const char **l, struct pz a[]);
int amp_rehost(struct abmail *m, int n, const char **l, struct pz a[]);
int amp_nabhost(struct abmail *m, int n, const char **l, struct pz a[]);
int amp_abhost(struct abmail *m, int n, const char **l, struct pz a[]);
int amp_helo(struct abmail *m, int n, const char **l, struct pz a[]);
int amp_boundary(struct abmail *m, int n, const char **l, struct pz a[]);
int amp_bomail(struct abmail *m, int n, const char **l, struct pz a[]);
int amp_bohelo(struct abmail *m, int n, const char **l, struct pz a[]);

#define RE_BYM "\\s+by\\s+.*"
#define RE_FRG "(?:\\s*\\(may be forged\\))?"
#define RE_ADR "\\[(.*)\\]"RE_FRG

struct headre {
  int sw;	/* switch on: !0, off: 0, loh on: >1  */
  int n;	/* nth header matching, n=-1 means for all the same headers */
  int neg;	/* negative matching on: !0, off: 0 */
  char *header;	/* header string */
  char *re;	/* regular expression */
  int (*proc)(struct abmail *, int, const char **, struct pz *);	/* process at matching */
  char *msg;	/* message string */
  int nv;	/* the number of vectors */
} headres[] = {
  { !0,  0,  0, "From ", "\\s*(([^\\s]*)@([^@\\s]*)|[^@\\s]*)", amp_bofrom, NULL },
  { !0,  0,  0, "Return-Path:", "\\s*<(([^\\s]*)@([^@\\s]*)|[^@\\s]*)>", amp_bofrom, NULL },
  { !0, -1,  0, "Received:", "(?:"RE_BYM"(?:\\s*.*)*|%s)\\s*$", amp_local, NULL },
  { !0,  0,  0, "Received:", "(?:%s)\\s*$", amp_auth, NULL },
  { !0,  0,  0, "Received:", "(?:%s)\\s*$", amp_subst, NULL, 1 },
  { !0,  0,  0, "Received:", "\\s*from\\s+(%s)\\s*\\(\\s*([^()\\s]*)\\s*"RE_ADR"\\)"RE_BYM, amp_abhelo, NULL },
  {  0,  0,  0, "Received:", "\\s*from\\s+(.*)\\s*\\(\\s*(unknown)\\s*"RE_ADR"\\)"RE_BYM, amp_rehost, NULL },
  { !0,  0,  0, "Received:", "\\s*from\\s+(.*)\\s*\\(\\s*(%s)\\s*"RE_ADR"\\)"RE_BYM, amp_nabhost, NULL },
  { !0,  0,  0, "Received:", "\\s*from\\s+(.*)\\s*\\(\\s*(%s)\\s*"RE_ADR"\\)"RE_BYM, amp_abhost, NULL },
  { !0,  0,  0, "Received:", "\\s*from\\s+(.*)\\s*\\(\\s*([^()\\s]*)\\s*"RE_ADR"\\)"RE_BYM, amp_helo, NULL },
  { !0,  0, !0, "From:", "\\s*.*", NULL, "without from header" },
  {  0,  0,  0, "To:", "\\s*([\\S\\s]*)\\s", amp_to, NULL },
  { !0,  0, !0, "Content-Type:", "\\s*multipart/(?:report);\\s*.*\\s*boundary=\\s*\"?([^\\s\";]*)", amp_boundary, NULL },
  {  2,  0,  0, "Content-Type:", "\\s*(message/rfc822|text/rfc822-headers)", amp_bomail, NULL },
  {  3, -1,  0, "Received:", "(?:"RE_BYM"(?:\\s*.*)*|%s)\\s*$", amp_local, NULL },
  {  3,  0,  0, "Received:", "(?:%s)\\s*$", amp_auth, NULL },
  {  3,  0,  0, "Received:", "(?:%s)\\s*$", amp_subst, NULL, 1 },
  {  3,  0,  0, "Received:", "\\s*from\\s+(%s)\\s*\\(\\s*([^()\\s]*)\\s*"RE_ADR"\\)"RE_BYM, amp_abhelo, NULL },
  {  3,  0,  0, "Received:", "\\s*from\\s+(.*)\\s*\\(\\s*(%s)\\s*"RE_ADR"\\)"RE_BYM, amp_nabhost, NULL },
  {  3,  0,  0, "Received:", "\\s*from\\s+(.*)\\s*\\(\\s*(%s)\\s*"RE_ADR"\\)"RE_BYM, amp_abhost, NULL },
  {  3,  0, !0, "Received:", "\\s*from\\s+(.*)\\s*\\(\\s*([^()\\s]*)\\s*"RE_ADR"\\)"RE_BYM, amp_bohelo, NULL },
};
enum {
  HR_bofrom,
  HR_borp,
  HR_local,
  HR_auth,
  HR_subst,
  HR_abhelo,
  HR_rehost,
  HR_nabhost,
  HR_abhost,
  HR_helo,
  HR_from,
  HR_to,
  HR_boundary,
  HR_bomail,
  HR_bolocal,
  HR_boauth,
  HR_bosubst,
  HR_boabhelo,
  HR_bonabhost,
  HR_boabhost,
  HR_bohelo,
};

int amp_to(struct abmail *m, int n, const char **l, struct pz a[])
{
  int rc = 0;

  if (n > 1 && l[1][0] == '\0')
    this_is_abmail("illegal to header for empty");
  return rc;
}

int amp_from(struct abmail *m, int n, const char **l, struct pz a[])
{
  int rc = 0, i;

  if (m->from[0][0] == '\0') {
    for (i=0; i<3; i++)
      if (n > i+1)
	strlcpy(m->from[i], l[i+1], sizeof(m->from[i]));
  }
  return rc;
}

int amp_bofrom(struct abmail *m, int n, const char **l, struct pz a[])
{
  int rc = 0, i;

  if (m->from[0][0] == '\0') {
    for (i=0; i<3; i++)
      if (n > i+1)
	strlcpy(m->from[i], l[i+1], sizeof(m->from[i]));
    if (m->from[0][0] == '\0' ||
	strncasecmp(m->from[0], "MAILER-DAEMON", sizeof("MAILER-DAEMON")-1) == 0) {
      m->body = !0;
    }
  }
  return rc;
}

int amp_local(struct abmail *m, int n, const char **l, struct pz a[])
{
  int rc = 0;
  char *p;

  if (n > 0) {
    if ((p=strchr(m->line, ':')))
      *p = ';';
  }
  return rc;
}

int amp_auth(struct abmail *m, int n, const char **l, struct pz a[])
{
  int rc = 0;

  if (n > 0) {
    this_is_not_abmail("message from authorized sender");
  }
  return rc;
}

int amp_subst(struct abmail *m, int n, const char **l, struct pz a[])
{
  int rc = 0, i, iov, irs;
  size_t j, j0, s;
  char line[sizeof(m->line)/sizeof(char)], *p;
  struct sd {
    char *src;
    size_t l;
    const char *dst;
  } *sds = NULL, sds0[] = {
    { strdup("\\n"),	sizeof("\\n")-1,	"\n" },
    { strdup("\\t"),	sizeof("\\t")-1,	"\t" },
    /* sds[] should be constructed as follows:
    { "\\10",		strlen("\\10"),		l[rss[irs].ub - 0] },
    { "\\9",		strlen("\\9"),		l[rss[irs].ub - 1] },
	:
    { "\\1",		strlen("\\1"),		l[rss[irs].ub - 9] },
     */
  };
  int nsd0 = sizeof(sds0)/sizeof(struct sd), nsd;
  const char *ses[] = {
    "host",
  };
  enum {
    SE_host,
  };

  for (iov=2; iov<n && a[iov].p == m->line-1; iov++);
  for (irs=0; irs<m->nrs && !(m->rss[irs].lb <= iov && iov <= m->rss[irs].ub); irs++);
  nsd = nsd0 + m->rss[irs].ub - m->rss[irs].lb + 1;

  if (!(sds = malloc(sizeof(struct sd)*nsd)))
    return rc;
  for (i=0; i<nsd; i++) {
    if (i < nsd0)
      sds[i] = sds0[i];
    else {
      int i0 = i - nsd0;
      char src[64];

      sprintf(src, "\\%d", m->rss[irs].ub - m->rss[irs].lb + 1 - i0);
      sds[i].src = strdup(src);
      sds[i].l = strlen(src);
      sds[i].dst = l[m->rss[irs].ub - i0];
    }
  }
  a[0].p[0] = '\0';
  strlcpy(line, m->line, sizeof(line)/sizeof(char));
  for (j0=0, j=0; j<(s=strlen(m->rss[irs].se)); j++) {
    if (!(p = strchr(&m->rss[irs].se[j], '\\'))) break;
    j = p - m->rss[irs].se;
    for (i=0; i<nsd; i++)
      if (strncmp(&m->rss[irs].se[j], sds[i].src, sds[i].l) == 0) {
	char buf[j-j0+1];

	strncpy(buf, &m->rss[irs].se[j0], j-j0); buf[j-j0] = '\0';
	strlcat(line, buf, sizeof(line)/sizeof(char));
	strlcat(line, sds[i].dst, sizeof(line)/sizeof(char));
	j += sds[i].l - 1;
	j0 = j + 1;
	break;
      }
  }
  if (j0 < s)
    strlcat(line, &m->rss[irs].se[j0], sizeof(line)/sizeof(char));
  for (i=0; i<nsd; i++)
    free(sds[i].src);
  free(sds);
  strlcpy(m->line, "", sizeof(m->line)/sizeof(char));
  for (j0=0, j=0; j<(s=strlen(line)); j++) {
    if (!(p = strchr(&line[j], '`'))) break;
    j = p - line;
    for (i=0; i<1; i++) {
      char src[64] = "`";
      size_t l;

      strlcat(src, ses[i], sizeof(src)/sizeof(char));
      strlcat(src, " ", sizeof(src)/sizeof(char));
      l = strlen(src);
      if (strncmp(&line[j], src, l) == 0 && (p = strstr(&line[j+l], "`"))) {
	char buf[j-j0+1];

	strncpy(buf, &line[j0], j-j0); buf[j-j0] = '\0';
	strlcat(m->line, buf, sizeof(m->line)/sizeof(char));
	switch (i) {
	case SE_host:
	  {
	    char address[64];
	    char h_name[2048];
	    int rv;

	    strlcpy(address, &line[j+l], min(p-&line[j+l]+1, sizeof(address)));
	    if ((rv=abadr_find(m, address, NULL)) == -1)
	      { this_is_abmail("illegal address"); return rc; }
	    else if (rv == 0)
	      strlcat(m->line, "known", sizeof(m->line)/sizeof(char));
	    else if (lookup_reverse(address, h_name, sizeof(h_name), m->paranoid))
	      strlcat(m->line, h_name, sizeof(m->line)/sizeof(char));
	    else
	      strlcat(m->line, "unknown", sizeof(m->line)/sizeof(char));
	  }
	  break;
	}
	j += l + ((p + sizeof("`")-1)-&line[j+l]) - 1;
	j0 = j + 1;
	break;
      }
    }
  }
  if (j0 < s)
    strlcat(m->line, &line[j0], sizeof(m->line)/sizeof(char));
  return rc;
}

#define fqdn l[1]
#define name l[2]
#define address l[3]
int amp_abhelo(struct abmail *m, int n, const char **l, struct pz a[])
{
  int rc, rv, c;

  if (n > 3) {
    if (m->verbose&(1<<3)) echo_abmail_helo();
    if ((rv=abadr_find(m, address, &c)) == -1)
      { this_is_abmail("illegal address"); return rc; }
    if (rv == 0) {
      if (c == -1) {
	this_is_not_abhelo("message from own address");
	return rc;
      }
      else if (c == 1 || c == 0) {
	this_is_abhelo("message from unreliable address or illegal received header");
	return rc;
      }
    }
  }
  this_is_abhelo("illegal received header");
  return rc;
}

int amp_rehost(struct abmail *m, int n, const char **l, struct pz a[])
{
  int rc = 0;
  struct in_addr in;
#if !defined(DISABLE_IPv6)
  struct in6_addr in6;
  int af = AF_INET;
#endif
  char h_name[2048];

  if (n > 3) {
    if (m->verbose&(1<<3)) echo_abmail_helo();
    if (inet_aton(address, &in) == 0)
#if !defined(DISABLE_IPv6)
      {
	if (inet_pton(AF_INET6, address, &in6) == 0)
	  { this_is_abhelo("illegal address"); return rc; }
	else
	  af = AF_INET6;
      }
#else
      { this_is_abhelo("illegal address"); return rc; }
#endif
    if (lookup_reverse(address, h_name, sizeof(h_name), m->paranoid)) {
      char line[sizeof(m->line)/sizeof(char)];

      a[2].p[0] = '\0';
      strlcpy(line, m->line, sizeof(line));
      strlcat(line, h_name, sizeof(line));
      strlcat(line, a[2].p + a[2].z, sizeof(line));
      strlcpy(m->line, line, sizeof(m->line));
    }
  }
  return rc;
}

int amp_nabhost(struct abmail *m, int n, const char **l, struct pz a[])
{
  int rc = 0;

  if (n > 0)
    if ((rc=amp_helo(m, n, l, a)) == 0)
      this_is_not_abhelo("message from reliable address for name");
  return rc;
}

int amp_abhost(struct abmail *m, int n, const char **l, struct pz a[])
{
  int rc, rv, c;

  if (n > 3) {
    if (m->verbose&(1<<3)) echo_abmail_helo();
    if ((rv=abadr_find(m, address, &c)) == -1)
      { this_is_abmail("illegal address"); return rc; }
    if (rv == 0) {
      if (c == -1 || c == 0) {
	this_is_not_abhelo("message from own or reliable address");
	return rc;
      }
      else if (c == 1) {
	this_is_abhelo("message from unreliable address");
	return rc;
      }
    }
  }
  this_is_abhelo("illegal received header for name");
  return rc;
}

int amp_helo(struct abmail *m, int n, const char **l, struct pz a[])
{
  int rc = 0, rv, c;

  if (n > 3) {
    if (m->verbose&(1<<3)) echo_abmail_helo();
    if ((rv=abadr_find(m, address, &c)) == -1)
      { this_is_abmail("illegal address"); return rc; }
    if (rv == 0) {
      if (c == -1 || c == 0) {
	this_is_not_abhelo("message from own or reliable address");
	return rc;
      }
      else if (c == 1) {
	this_is_abhelo("message from unreliable address");
	return rc;
      }
    }
    if (!strchr(fqdn, '.')) {
      this_is_abhelo("illegal received header for fqdn");
      return rc;
    }
    if (m->dnsbl) {
      int i;

      if ((i = dnsbl_find(address))) {
	this_is_abhelo("message from wrong address for dns black list");
	if (m->verbose&(1<<0)) echo_abmail_dnsbl();
      }
    }
  }
  return rc;
}

int amp_boundary(struct abmail *m, int n, const char **l, struct pz a[])
{
  int rc = 0;

  if (n > 1) {
    if (m->lom < sizeof(m->boundary)/sizeof(m->boundary[0])) {
      m->lom++;
      strlcpy(m->boundary[m->lom-1], "--", sizeof(m->boundary[m->lom-1]));
      strlcat(m->boundary[m->lom-1], l[1], sizeof(m->boundary[m->lom-1]));
    }
  }
  else
    if (m->body) this_is_abmail("illegal multipart");
  return rc;
}

int amp_bomail(struct abmail *m, int n, const char **l, struct pz a[])
{
  m->hop = !0;
  return 0;
}

int amp_bohelo(struct abmail *m, int n, const char **l, struct pz a[])
{
  int rc = 0;

  if (n > 0)
    rc = amp_helo(m, n, l, a);
  else
    this_is_abhelo("illegal bounced message");
  return rc;
}
#undef fqdn
#undef name
#undef address

void res_read(int argc, char *argv[])
{
  char *myname = strdup(argv[0]), *herename = dirname(myname);
  char buf[2048], lst_pathname[MAXPATHLEN], *lst_basename[] = {
    "rfs_reliable.lst",
    "ras_reliable.lst",
    "rss_reliable.lst",
    "rhs_unreliable.lst",
    "rns_reliable.lst",
    "rns_unreliable.lst",
  };
  enum {
    LST_rfs_reliable,
    LST_ras_reliable,
    LST_rss_reliable,
    LST_rhs_unreliable,
    LST_rns_reliable,
    LST_rns_unreliable,
  };
  int i, j = 0, n = 1;
  FILE *fp;
  char *p;
  char *re;
  struct abmail *m = &me;

  for (i=0; i<sizeof(lst_basename)/sizeof(char *); i++) {
    int nv = 1;

    re = NULL;
    sprintf(lst_pathname, "%s/%s", herename, lst_basename[i]);
    if (!(fp=fopen(lst_pathname, "r"))) {
      char *home = getenv("HOME");
      sprintf(lst_pathname, "%s/.abmail/%s", home?home:"", lst_basename[i]);
      if (!(fp=fopen(lst_pathname, "r"))) {
	sprintf(lst_pathname, PREFIX"/etc/abmail/%s", lst_basename[i]);
	if (!(fp=fopen(lst_pathname, "r")))
	  ;
      }
    }
    switch (i) {
    case LST_rss_reliable: n = 2; break;
    default: n = 1; break;
    }
    while (fp && fgets(buf, sizeof(buf), fp)) {
      int c = 0, d = 0;

      if (buf[0] == '#' || buf[0] == '\n') continue;
      p = strtok(buf, "\t\n");
      while (p) {
	switch (c) {
	case 0:
	  if (!re) {
	    re = malloc(sizeof(char)*(strlen(p)+1));
	    if (re) re[0] = '\0';
	  }
	  else {
	    re = realloc(re, sizeof(char)*(strlen(re)+strlen("|")+strlen(p)+1));
	    if (re) strcat(re, "|");
	  }
	  if (re) strcat(re, p);
	  switch (i) {
	  case LST_rss_reliable:
	    for (j=0; j<strlen(p); j++) {
	      if (p[j] == '(') d++;
	      if (strncmp(&p[j], "\\(", sizeof("\\(")-1) == 0) d--;
	      if (strncmp(&p[j], "(?:", sizeof("(?:")-1) == 0) d--;
	    }
	    m->rss = m->rss?realloc(m->rss, sizeof(struct rs)*(m->nrs+1)):
	      malloc(sizeof(struct rs)*(m->nrs+1));
	    if (m->rss) {
	      m->rss[m->nrs].lb = nv;
	      m->rss[m->nrs].ub = nv + d - 1;
	    }
	    nv += d;
	    break;
	  }
	  break;
	case 1:
	  switch (i) {
	  case LST_rss_reliable:
	    if (m->rss) m->rss[m->nrs].se = strdup(p);
	    m->nrs++;
	    break;
	  }
	  break;
	}
	p = strtok(NULL, "\t\n");
	if (++c == n) break;
      }
    }
    if (fp) fclose(fp);
    switch (i) {
    case LST_rfs_reliable: j = HR_local; break;
    case LST_ras_reliable: j = HR_auth; break;
    case LST_rss_reliable: j = HR_subst; headres[j].nv = nv; break;
    case LST_rhs_unreliable: j = HR_abhelo; break;
    case LST_rns_reliable: j = HR_nabhost; break;
    case LST_rns_unreliable: j = HR_abhost; break;
    }
    if ((p=malloc(sizeof(char)*(strlen(headres[j].re)-2+strlen(re?re:"")+1)))) {
      sprintf(p, headres[j].re, re?re:"");
      headres[j].re = p;
    }
    switch (i) {
    case LST_rfs_reliable: j = HR_bolocal; break;
    case LST_ras_reliable: j = HR_boauth; break;
    case LST_rss_reliable: j = HR_bosubst; headres[j].nv = nv; break;
    case LST_rhs_unreliable: j = HR_boabhelo; break;
    case LST_rns_reliable: j = HR_bonabhost; break;
    case LST_rns_unreliable: j = HR_boabhost; break;
    }
    if ((p=malloc(sizeof(char)*(strlen(headres[j].re)-2+strlen(re?re:"")+1)))) {
      sprintf(p, headres[j].re, re?re:"");
      headres[j].re = p;
    }
    if (re) free(re);
  }
  if (myname) free(myname);
}

struct matching {
  pcre *re;
  const char *errmsg;
  int erroffset;
  int *ovec;
  int rc;
};
struct process_headre {
  struct headre *h;
  int c, flag;
  struct matching m;
} process_headres[sizeof(headres)/sizeof(struct headre)];

int abfscan(struct abmail *m, struct process_headre *process_headres, FILE *fp)
{
  int rc = 0, eoh = 0, bob = 0, i;

  strlcpy(m->line, "", sizeof(m->line));
  while (fgets(m->buf, sizeof(m->buf), fp) != NULL && (!eoh || m->lom)) {
    if (!eoh) {
      size_t l;

      if ((l=strlen(m->buf)) == 0 || m->buf[l-1] != '\n') {
	this_is_abmail("illegal header");
	break;
      }
      if (m->buf[0] == '\n')
	eoh = !0;
      else if (isspace((int)m->buf[0])) {
	strlcat(m->line, m->buf, sizeof(m->line));
	continue;
      }
    }
    else {
      if (!m->body) break;
      if (m->lom && m->boundary[m->lom-1][0] != '\0') {
	if (strncmp(m->buf, m->boundary[m->lom-1], strlen(m->boundary[m->lom-1])) != 0) {
	  continue;
	}
	else {
	  if (strncmp(&m->buf[strlen(m->boundary[m->lom-1])], "--", 2) != 0) {
	    eoh = bob = 0;
	    if (m->loh == 0)
	      m->loh = 1;
	    continue;
	  } else
	    if (!--m->lom)
	      break;
	}
      }
      strlcpy(m->line, m->buf, sizeof(m->line));
    }
    for (i=0; i<sizeof(headres)/sizeof(struct headre) && rc == 0; i++) {
      struct process_headre *p = &process_headres[i];

      if (p->h->sw - m->loh != 1) continue;
      if (!p->m.re) continue;
      if (strncasecmp(p->h->header, m->line, strlen(p->h->header)) == 0) {
	if (p->h->n < 0 || (p->h->n == p->c)) {
	  p->m.rc = pcre_exec(p->m.re, NULL,
			      m->line, sizeof(m->line), strlen(p->h->header),
			      0, p->m.ovec, p->m.ovec ? p->h->nv*3 : 0);
	  if (p->m.rc >= 0) {
	    if (!p->h->proc) {
	      if (!p->h->neg)
		this_is_abmail(p->h->msg);
	    }
	    else {
	      const char **l;
	      struct pz a[p->h->nv];

	      if (!pcre_get_substring_list(m->line, p->m.ovec, p->m.rc, &l)) {
		pcre_set_substring_pz_array(m->line, p->m.ovec, p->m.rc, a);
		rc = p->h->proc(m, p->m.rc, l, a);
		pcre_free_substring_list(l);
	      }
	    }
	    if (p->h->neg)
	      p->flag = 0;
	  }
	}
	p->c++;
      }
    }
    if (!m->lom && rc != 0) break;
    if (eoh && !bob) {
      for (i=0; i<sizeof(headres)/sizeof(struct headre) && rc == 0; i++) {
	struct process_headre *p = &process_headres[i];

	p->c = 0;
	if (p->h->sw - m->loh != 1) continue;
	if (!p->m.re) continue;
	if (!p->h->neg) continue;
	if (p->flag) {
	  if (!p->h->proc)
	    this_is_abmail(p->h->msg);
	  else
	    rc = p->h->proc(m, 0, NULL, NULL);
	}
      }
      bob = !0;
      if (m->loh == 1 && m->hop) {
	eoh = bob = 0;
	m->loh = 2;
      }
      else if (m->loh == 2) {
	m->lom = m->hop = 0;
	m->loh = 1;
      }
    }
    strlcpy(m->line, m->buf, sizeof(m->line));
  }
  if (m->hop) this_is_abmail("unexpected report format");
  this_was_not_abmail("no doubts in this message");
  for (i=0; i<sizeof(headres)/sizeof(struct headre); i++) {
    struct process_headre *p = &process_headres[i];

    p->c = 0;
    p->flag = !0;
  }
  return rc;
}

#if !defined(DISABLE_MMAN)
#define SM_PATHNAME "/var/tmp/org.abmail.abmail"
#if !defined(DISABLE_SHM)
#define sm_open(name, flags, mode) shm_open(name, flags, mode)
#define sm_unlink(name) shm_unlink(name)
#else
#define sm_open(path, flags, mode) open(path, flags, mode)
#define sm_unlink(path) unlink(path)
#endif
struct sm_tag {
  size_t z, nrs, nam, nam6;
  struct item_headre {
    int nv;
    size_t z;
  } item_headres[sizeof(headres)/sizeof(struct headre)];
} sm_tag = { 0, 0, 0, 0, };

int sm_obtain(const char *name)
{
  int rc = 0, fd, i;
  void *m;
  size_t z;

  if ((fd=sm_open(name, O_RDONLY, 0644)) &&
      (m=mmap(0, sizeof(sm_tag), PROT_READ, MAP_SHARED, fd, 0)) != (void *)-1 &&
      (z=((struct sm_tag *)m)->z) &&
      !munmap(m, sizeof(sm_tag)) &&
      (m=mmap(0, z, PROT_READ, MAP_SHARED, fd, 0)) != (void *)-1) {
    struct sm_tag *tag;

    close(fd);
    tag = (struct sm_tag *)m;
    m += sizeof(struct sm_tag);
    memcpy(am_masks, m, sizeof(am_masks));
    m += sizeof(am_masks);
#if !defined(DISABLE_IPv6)
    memcpy(am6_masks, m, sizeof(am6_masks));
    m += sizeof(am6_masks);
#endif
    if ((me.rss=malloc(sizeof(struct rs)*tag->nrs)))
      memcpy(me.rss, m, sizeof(struct rs)*tag->nrs);
    m += sizeof(struct rs)*tag->nrs;
    me.as->ams = (struct am *)m;
    m += sizeof(struct am)*tag->nam;
#if !defined(DISABLE_IPv6)
    me.a6s->ams = (struct am6 *)m;
    m += sizeof(struct am6)*tag->nam6;
#endif
    for (i=0; i<sizeof(headres)/sizeof(struct headre); i++) {
      struct process_headre *p = &process_headres[i];

      p->h = &headres[i];
      p->m.re = tag->item_headres[i].z ? (pcre *)m : NULL;
      m += tag->item_headres[i].z;
      p->h->nv = tag->item_headres[i].nv;
    }
    for (i=0; i<tag->nrs; i++) {
      me.rss[i].se = m;
      m += strlen(me.rss[i].se) + 1;
    }
    me.nrs = tag->nrs;
    me.as->nam = tag->nam;
#if !defined(DISABLE_IPv6)
    me.a6s->nam = tag->nam6;
#endif
  }
  else
    rc = 1;
  return rc;
}

int sm_store(const char *name)
{
  int rc = 0, fd, i;
  size_t z = sizeof(sm_tag), l;
  void *m;

  z += sizeof(am_masks);
#if !defined(DISABLE_IPv6)
  z += sizeof(am6_masks);
#endif
  z += sizeof(struct rs)*(sm_tag.nrs=me.nrs);
  z += sizeof(struct am)*(sm_tag.nam=me.as->nam);
#if !defined(DISABLE_IPv6)
  z += sizeof(struct am6)*(sm_tag.nam6=me.a6s->nam);
#endif
  for (i=0; i<sizeof(headres)/sizeof(struct headre); i++)
    z += sm_tag.item_headres[i].z;
  for (i=0; i<sm_tag.nrs; i++)
    z += (me.rss[i].se ? strlen(me.rss[i].se) : 0) + 1;
  sm_unlink(name);
  if ((fd=sm_open(name, O_CREAT|O_EXCL|O_RDWR, 0644)) &&
      ftruncate(fd, z) == 0 &&
      (m=mmap(0, z, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0)) != (void *)-1) {
    struct sm_tag *tag = (struct sm_tag *)m;

    close(fd);
    memcpy(m, &sm_tag, sizeof(struct sm_tag));
    m += sizeof(struct sm_tag);
    memcpy(m, am_masks, sizeof(am_masks));
    m += sizeof(am_masks);
#if !defined(DISABLE_IPv6)
    memcpy(m, am6_masks, sizeof(am6_masks));
    m += sizeof(am6_masks);
#endif
    memcpy(m, me.rss, sizeof(struct rs)*sm_tag.nrs);
    m += sizeof(struct rs)*sm_tag.nrs;
    for (i=0; i<sm_tag.nam; i++)
      me.as->ams[i].c = NULL;
    memcpy(m, me.as->ams, sizeof(struct am)*sm_tag.nam);
    m += sizeof(struct am)*sm_tag.nam;
#if !defined(DISABLE_IPv6)
    for (i=0; i<sm_tag.nam6; i++)
      me.a6s->ams[i].c = NULL;
    memcpy(m, me.a6s->ams, sizeof(struct am6)*sm_tag.nam6);
    m += sizeof(struct am6)*sm_tag.nam6;
#endif
    for (i=0; i<sizeof(headres)/sizeof(struct headre); i++) {
      struct process_headre *p = &process_headres[i];

      if (sm_tag.item_headres[i].z)
	memcpy(m, p->m.re, sm_tag.item_headres[i].z);
      m += sm_tag.item_headres[i].z;
    }
    for (i=0; i<sm_tag.nrs; i++) {
      memcpy(m, me.rss[i].se?me.rss[i].se:"",
	     (l=me.rss[i].se?strlen(me.rss[i].se):0) + 1);
      m += l + 1;
    }
    tag->z = z;
  }
  else
    rc = 1;
  return rc;
}
#endif

#if !defined(DISABLE_PTHREAD)
struct thr_abfscans_arg {
  char *pathname;
  struct abmail *m;
  int rc;
};

int thr_abfscans_enable = 0, m_thr = FOPEN_MAX;

void *thr_abfscans(void *arg)
{
  struct thr_abfscans_arg *a = (struct thr_abfscans_arg *)arg;
  FILE *fp;

  if (!(fp = fopen(a->pathname, "r"))) {
    free(a->pathname);
    a->pathname = NULL;
    return NULL;
  }
  {
    int i;
    struct process_headre phs[sizeof(headres)/sizeof(struct headre)];

    memcpy(phs, process_headres, sizeof(process_headres));
    for (i=0; i<sizeof(headres)/sizeof(struct headre); i++) {
      struct process_headre *p = &phs[i];

      p->m.ovec = malloc(sizeof(int)*p->h->nv*3);
    }      
    a->rc = abfscan(a->m, phs, fp);
    for (i=0; i<sizeof(headres)/sizeof(struct headre); i++) {
      struct process_headre *p = &phs[i];

      if (p->m.ovec) free(p->m.ovec);
    }
  }
  fclose(fp);
  return NULL;
}
#endif

char *mailpath(const char *pathname, size_t *sz, int *mailpathtype)
{
  int i, is_maildir = 0;
  char *rv, *maildirs[] = { "cur", "tmp", "new" }; /* last should be "new" */
  size_t l = pathname ? strlen(pathname) : 0,
    z = (l ? l+1 : 0) + max(max(sizeof("cur"), sizeof("tmp")), sizeof("new"));
  struct stat sb;

  if (sz)
    *sz = l+1;
  if ((rv = malloc(z))) {
    for (i=0; i<sizeof(maildirs)/sizeof(char *); i++) {
      strcpy(rv, "");
      if (l) {
	strcat(rv, pathname);
	strcat(rv, "/");
      }
      strcat(rv, maildirs[i]);
      if (!(is_maildir=(stat(rv, &sb) == 0 && S_ISDIR(sb.st_mode))))
	break;
    }
    if (is_maildir)
      *sz = z;
    else {
      free(rv);
      rv = pathname ? strdup(pathname) : NULL;
    }      
  }
  if (mailpathtype)
    *mailpathtype = is_maildir;
  return rv;
}

int main(int argc, char *argv[])
{
  int rc = 0, i, a;
  int rescue = 0, greed = 0, bound = 0, attach = 0, debug = 0, about = !0;
  FILE *fp = stdin;
  int sm_server = 0, sm_client = 0;
#if !defined(DISABLE_MMAN)
  char sm_pathname[MAXPATHLEN] = SM_PATHNAME;
#define set_sm_pathname(opt) do {\
  char *p = argv[a] + sizeof(opt)-1;\
  size_t l;\
\
  if (*p == '=' && (l=strlen(p+1)) && l < sizeof(sm_pathname))\
    strlcpy(sm_pathname, p+1, sizeof(sm_pathname));\
  else if (*p != '\0') break;\
  if (debug) fprintf(stdout, "abmail: sm_pathname: %s\n", sm_pathname);\
} while (0)
#endif
#if !defined(DISABLE_PTHREAD)
  int n_thr = 0;
  struct thr_abfscans_arg *thr_abfscans_args = NULL;
#define do_thr_abfscans() do {\
  int i, j;\
  struct abmail *thr_ms;\
\
  if ((thr_ms = malloc(sizeof(struct abmail)*min(n_thr, m_thr)))) {\
    for (i=0; i<n_thr; i+=min(n_thr, m_thr)) {\
      pthread_t threads[min(n_thr, m_thr)];\
      int rc[min(n_thr, m_thr)];\
      struct thr_abfscans_arg *args = &thr_abfscans_args[i];\
\
      for (j=0; j<min(n_thr, m_thr) && i+j<n_thr; j++) {\
	thr_ms[j] = me;\
	args[j].m = &thr_ms[j];\
	rc[j] = pthread_create(&threads[j], NULL, thr_abfscans, (void *)&args[j]);\
      }\
      for (j=0; j<min(n_thr, m_thr) && i+j<n_thr; j++) {\
	if (rc[j] == 0)\
	  pthread_join(threads[j], NULL);\
	else\
	  thr_abfscans((void *)&args[j]);\
	if (args[j].pathname) {\
	  if ((about && args[j].rc == 1) || (!about && args[j].rc != 1))\
	    printf("%s\n", args[j].pathname);\
	  free(args[j].pathname);\
	  args[j].pathname = NULL;\
	}\
      }\
    }\
    free(thr_ms);\
  }\
  else\
    for (i=0; i<n_thr; i+=min(n_thr, m_thr)) {\
      struct thr_abfscans_arg *args = &thr_abfscans_args[i];\
\
      for (j=0; j<min(n_thr, m_thr) && i+j<n_thr; j++) {\
	struct abmail m = me;\
\
	args[j].m = &m;\
	thr_abfscans((void *)&args[j]);\
	if (args[j].pathname) {\
	  if ((about && args[j].rc == 1) || (!about && args[j].rc != 1))\
	    printf("%s\n", args[j].pathname);\
	  free(args[j].pathname);\
	  args[j].pathname = NULL;\
	}\
      }\
    }\
  if (thr_abfscans_args) free(thr_abfscans_args);\
  thr_abfscans_args = NULL;\
  n_thr = 0;\
} while (0)
#endif

  me.msgout = stderr;	/* ! */
  for (a=1; a<argc; a++)
    if (strcmp(argv[a], "--") == 0)
      break;
    else if (strcmp(argv[a], "-D") == 0)
      debug = 1;
    else if (strncmp(argv[a], "--sm-server", sizeof("--sm-server")-1) == 0)
      {
#if !defined(DISABLE_MMAN)
	set_sm_pathname("--sm-server");
	sm_server = !0;
#endif
      }
    else if (strncmp(argv[a], "--sm-delete", sizeof("--sm-delete")-1) == 0)
      {
#if !defined(DISABLE_MMAN)
	set_sm_pathname("--sm-delete");
	return sm_unlink(sm_pathname);
#endif
      }
    else if (strncmp(argv[a], "--sm-client", sizeof("--sm-client")-1) == 0)
      {
#if !defined(DISABLE_MMAN)
	set_sm_pathname("--sm-client");
	sm_client = !0;
#endif
      }
    else if (strncmp(argv[a], "--am-daemon=", sizeof("--am-daemon=")-1) == 0) {
#if defined(USE_ABADDR_DB)
      me.host = argv[a] + sizeof("--am-daemon=")-1;
#endif
    }
    else if (strncmp(argv[a], "--am-dbfile=", sizeof("--am-dbfile=")-1) == 0) {
#if defined(USE_ABADDR_DB)
      me.file = argv[a] + sizeof("--am-dbfile=")-1;
#endif
    }
    else if (strcmp(argv[a], "-M-") == 0)
#if !defined(DISABLE_PTHREAD)
      thr_dnsbl_find_enable = 0
#endif
	;
    else if (strcmp(argv[a], "-MM") == 0)
#if !defined(DISABLE_PTHREAD)
      thr_abfscans_enable = !0
#endif
	;
    else if (strncmp(argv[a], "-MM=", sizeof("-MM=")-1) == 0)
      {
#if !defined(DISABLE_PTHREAD)
	char *p = argv[a] + sizeof("-MM=")-1;
	int n = atoi(p);

	thr_abfscans_enable = !0;
	if (0 < n)
	  m_thr = n;
#endif
      }
    else if (strcmp(argv[a], "-MM-") == 0)
#if !defined(DISABLE_PTHREAD)
      thr_abfscans_enable = 0
#endif
	;
    else if (strcmp(argv[a], "-P-") == 0)
      me.paranoid = 0;
    else if (strcmp(argv[a], "-c") == 0)
      me.msgout = stdout;
    else if (strcmp(argv[a], "-vvvv") == 0)
      me.verbose |= (1<<3);
    else if (strcmp(argv[a], "-vvv") == 0)
      me.verbose |= (1<<2);
    else if (strcmp(argv[a], "-vv") == 0)
      me.verbose |= (1<<0)|(1<<1);
    else if (strcmp(argv[a], "-v-") == 0)
      me.verbose |= (1<<1);
    else if (strcmp(argv[a], "-v") == 0)
      me.verbose |= (1<<0);
    else if (strcmp(argv[a], "-d") == 0)
      me.dnsbl = 1;
    else if (strcmp(argv[a], "-r") == 0)
      rescue = 1;
    else if (strcmp(argv[a], "-g") == 0)
      greed = 1;
    else if (strcmp(argv[a], "-b") == 0)
      bound = 1;
    else if (strcmp(argv[a], "-a") == 0)
      { bound = 1; attach = 1; }
    else if (strcmp(argv[a], "-t") == 0)
      headres[HR_to].sw = !0;
    else if (strcmp(argv[a], "+") == 0)
      { fp = NULL; about = !0; }
    else if (strcmp(argv[a], "-") == 0)
      { fp = NULL; about = 0; }
    else
      break;
#if !defined(DISABLE_MMAN)
  if (sm_client) {
    sm_client = (sm_obtain(sm_pathname) == 0);
    if (!sm_client && debug)
      fprintf(stdout, "abmail: sm-client failed for %s\n", sm_pathname);
  }
#endif
  if (!sm_client) {
    am_init_masks();
#if !defined(DISABLE_IPv6)
    am6_init_masks();
#endif
    amset_read(me.as,
#if !defined(DISABLE_IPv6)
	       me.a6s,
#endif
	       0, debug, argc, argv);
    res_read(argc, argv);
  }
  if (rescue == 1)
    headres[HR_rehost].sw = !0;
  if (greed == 0) {
    if (!bound) headres[HR_bofrom].sw = headres[HR_borp].sw = 0;
    headres[HR_abhost].sw = headres[HR_helo].sw =
      headres[HR_boabhost].sw = headres[HR_bohelo].sw = 0;
    if (me.dnsbl) headres[HR_helo].sw = headres[HR_bohelo].sw = !0;
    /*if (!sm_client) me.as->nam = sizeof(builtin_ams)/sizeof(struct am);*/
  }
  if (bound == 0) {
    headres[HR_bofrom].proc = headres[HR_borp].proc = amp_from;
    headres[HR_boundary].sw = headres[HR_bomail].sw =
      headres[HR_bolocal].sw = headres[HR_boauth].sw = headres[HR_bosubst].sw =
      headres[HR_boabhelo].sw = headres[HR_bonabhost].sw =
      headres[HR_boabhost].sw = headres[HR_bohelo].sw = 0;
  }
  if (attach == 1) {
    headres[HR_bofrom].header = "From: ";
    headres[HR_borp].sw =
      headres[HR_local].sw = headres[HR_auth].sw = headres[HR_subst].sw =
      headres[HR_abhelo].sw = headres[HR_rehost].sw = headres[HR_nabhost].sw =
      headres[HR_abhost].sw = headres[HR_helo].sw =
      headres[HR_from].sw = headres[HR_to].sw = 0;
  }
  if (sm_client && debug) {
    am_check(me.as->nam, me.as->ams);
#if !defined(DISABLE_IPv6)
    am6_check(me.a6s->nam, me.a6s->ams);
#endif
  }

  for (i=0; i<sizeof(headres)/sizeof(struct headre); i++) {
    struct process_headre *p = &process_headres[i];

    p->h = &headres[i];
    p->c = 0;
    p->flag = !0;
    if (!sm_server && p->h->sw == 0) continue;
    if (!sm_client) {
    p->m.re = pcre_compile(p->h->re,
			   PCRE_ANCHORED|PCRE_CASELESS,
			   &p->m.errmsg, &p->m.erroffset, NULL);
    if (!p->m.re)
      fprintf(stderr, "pcre compilation failed at offset %d: %s\n%s\n",
	      p->m.erroffset, p->m.errmsg, p->h->re);
    }
    p->h->nv = max(10, p->h->nv);
#if !defined(DISABLE_MMAN)
    if (sm_server) {
      if (p->m.re)
	pcre_fullinfo(p->m.re, NULL, PCRE_INFO_SIZE, &sm_tag.item_headres[i].z);
      sm_tag.item_headres[i].nv = p->h->nv;
    }
#endif
    p->m.ovec = malloc(sizeof(int)*p->h->nv*3);
  }
#if !defined(DISABLE_MMAN)
  if (sm_server) {
    if ((rc=sm_store(sm_pathname)) != 0)
      fprintf(stderr, "abmail: sm-server failed for %s\n", sm_pathname);
    return rc;
  }
#endif
  if (fp) {
    rc = abfscan(&me, process_headres, fp);
  }
  else {
    for (; a<argc; a++) {
      int rc = 0;
      struct stat sb;

      if (stat(argv[a], &sb) != 0)
	continue;
      if (S_ISDIR(sb.st_mode)) {
	DIR *dir;
	struct dirent *d;
	size_t z;
	char *mailpathname;
	int mailpathtype;

	mailpathname = mailpath(argv[a], &z, &mailpathtype);
	if (!(dir = opendir(mailpathname))) {
	  if (mailpathname) free(mailpathname);
	  continue;
	}
	while ((d = readdir(dir))) { /* ! */
	  char *pathname;

	  if (!mailpathtype && !aredigit(d->d_name))
	    continue;
	  if ((pathname = malloc(z + 1 + strlen(d->d_name)))) {
	    strcpy(pathname, mailpathname);
	    strcat(pathname, "/");
	    strcat(pathname, d->d_name);
	  }
	  if (stat(pathname, &sb) == 0 && S_ISDIR(sb.st_mode)) {
	    if (pathname) free(pathname);
	    continue;
	  }
#if !defined(DISABLE_PTHREAD)
	  if (!thr_abfscans_enable) {
#endif
          {
	    struct abmail m = me;

	    if (!(fp = fopen(pathname, "r"))) {
	      if (pathname) free(pathname);
	      continue;
	    }
	    rc = abfscan(&m, process_headres, fp);
	    if ((about && rc == 1) || (!about && rc != 1))
	      printf("%s\n", pathname);
	    fclose(fp);
	  }
	  if (pathname) free(pathname);
#if !defined(DISABLE_PTHREAD)
	  }
	  else {
	    thr_abfscans_args = thr_abfscans_args?
	      realloc(thr_abfscans_args, sizeof(struct thr_abfscans_arg)*(n_thr+1)):
	      malloc(sizeof(struct thr_abfscans_arg)*(n_thr+1));
	    if (thr_abfscans_args)
	      thr_abfscans_args[n_thr].pathname = pathname;
	    n_thr++;
	    if (n_thr == m_thr) do_thr_abfscans();
	  }
#endif
	}
	closedir(dir);
	if (mailpathname) free(mailpathname);
	continue;
      }
#if !defined(DISABLE_PTHREAD)
      if (!thr_abfscans_enable) {
#endif
      {
	struct abmail m = me;

	if (!(fp = fopen(argv[a], "r")))
	  continue;
	rc = abfscan(&m, process_headres, fp);
	if ((about && rc == 1) || (!about && rc != 1))
	  printf("%s\n", argv[a]);
	fclose(fp);
      }
#if !defined(DISABLE_PTHREAD)
      }
      else {
	thr_abfscans_args = thr_abfscans_args?
	  realloc(thr_abfscans_args, sizeof(struct thr_abfscans_arg)*(n_thr+1)):
	  malloc(sizeof(struct thr_abfscans_arg)*(n_thr+1));
	if (thr_abfscans_args)
	  thr_abfscans_args[n_thr].pathname = strdup(argv[a]);
	n_thr++;
	if (n_thr == m_thr) do_thr_abfscans();
      }
#endif
    }
#if !defined(DISABLE_PTHREAD)
    if (!thr_abfscans_enable) ;
    else do_thr_abfscans();
#endif
  }
  return rc;
}
