/*
 * abmisc.c - abmail miscellaneous routines
 *
 * Copyright (C) 2006-2008 Taiji Yamada <taiji@aihara.co.jp>
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Affero General Public License for more details.
 */
#include <stdlib.h>	/* qsort, bsearch, etc */
#include <string.h>	/* strcmp, etc */
#include <libgen.h>	/* dirname, etc */
#include <arpa/inet.h>	/* inet_ntoa, etc */
#include <sys/types.h>	/* sys/socket.h */
#include <sys/socket.h>	/* AF_INET, etc */
#include <netdb.h>	/* gethostbyname, etc */
#include <unistd.h>	/* gethostname, etc */
#include <sys/param.h>	/* MAXPATHLEN, etc */
#include "abcidr.h"
#include "abmisc.h"

#if !defined(MYNETWORK)
#define MYNETWORK "210.154.62.64"
//#define MYNETMASK "255.255.255.248"	/* mask */
#define MYNETMASK "29"			/* or cidr */
//#define MYNETMASK "-210.154.62.71"	/* or broadcast */
#define MYDOMAIN "aihara.co.jp"
#endif

#if !defined(HAVE_STRLCPY)
#define strlcpy(d, s, z) do { strncpy(d, s, z); if ((z) > 0) (d)[(z)-1] = '\0'; } while (0)
#endif

#if !defined(HAVE_STRLCAT)
#define strlcat(d, s, z) do { if ((z) > strlen(d)) strncat(d, s, (z)-strlen(d)-1); } while (0)
#endif

#if !defined(HAVE_INET_ATON)
#define inet_aton(s, p) ((((p)->s_addr=inet_addr(s))!=-1||strcmp(s, "255.255.255.255")==0)?1:0)
#endif

#if !defined(PREFIX)
#define PREFIX "/opt/local"
#endif

struct am default_ams_clients[] = {
  { "127.0.0.0",	/* 127.255.255.255 */	"8",	-1,	"LOOPBACK"	},
};

struct am builtin_ams[] = {
  /* local private and global addresses */
  { "127.0.0.0",	/* 127.255.255.255 */	"8",	-1,	"LOOPBACK"	},
  { "10.0.0.0",		/* 10.255.255.255 */	"8",	-1,	"PRIVATE-A"	},
  { "172.16.0.0",	/* 172.31.255.255 */	"12",	-1,	"PRIVATE-B"	},
  { "192.168.0.0",	/* 192.168.255.255 */	"16",	-1,	"PRIVATE-C"	},
  { MYNETWORK,		MYNETMASK,			-1,	MYDOMAIN	},
  /* reliable global addresses */
#if defined(USE_AMS_RELIABLE_H)
#include "ams_reliable.h"
#endif
  /* unreliable global addresses */
#if defined(USE_AMS_UNRELIABLE_H)
#include "ams_unreliable.h"
#endif
};

struct amset amset_clients = { 0, NULL, NULL }, amset = {
  sizeof(builtin_ams)/sizeof(builtin_ams[0]), builtin_ams, builtin_ams
};

#define USE_BSEARCH_THRESHOLD_N 32

struct am *am_bsearch(int n, struct am ams[], struct in_addr a)
{
  struct am am_k;

  am_k.ads[0] = a;
  am_k.ads[1].s_addr = -1;
  am_k.ads[2] = a;
  am_k.ads[3] = a;
  return bsearch(&am_k, ams, n, sizeof(struct am), am_compare);
}
struct am *am_lsearch(int n, struct am ams[], struct in_addr a)
{
  int i;

  for (i=0; i<n; i++)
    if (ntohl(ams[i].ads[2].s_addr) <= ntohl(a.s_addr) &&
	ntohl(a.s_addr) <= ntohl(ams[i].ads[3].s_addr))
      break;
  return (i!=n)?&ams[i]:NULL;
}
struct am *am_search(int n, struct am ams[], struct in_addr a)
{
  struct am *rv = NULL;

  if (n > USE_BSEARCH_THRESHOLD_N)
    rv = am_bsearch(n, ams, a);
  else
    rv = am_lsearch(n, ams, a);
  return rv;
}

void am_set_spaces(size_t nam, struct am ams[])
{
  int i;

  for (i=0; i<nam; i++) {
    if (inet_aton(ams[i].a, &ams[i].ads[0]) == 0)
      fprintf(stderr, "inet_aton failed for %s\n", ams[i].a);
    am_set_space(&ams[i]);
  }
}

void am_check(size_t nam, struct am ams[])
{
  int i;

  for (i=0; i<nam; i++) {
    if (ams[i].m[0] != '-')
      fprintf(stdout, "  { \"%s\",\t/* %s */\t\"%s\",\t%d,\t\"%s\"\t},",
	      ams[i].a, inet_ntoa(ams[i].ads[3]), ams[i].m, ams[i].rc, ams[i].c?ams[i].c:"");
    else
      fprintf(stdout, "  { \"%s\",\t\"%s\",\t\t%d,\t\"%s\"\t},",
	      ams[i].a, ams[i].m, ams[i].rc, ams[i].c?ams[i].c:"");
    if (i > 0 &&
	(ams[i-1].rc == ams[i].rc &&
	 htonl(ntohl(ams[i-1].ads[3].s_addr) + 1) == ams[i].ads[2].s_addr))
      fprintf(stdout, "\t/* may connect these range */");
    fprintf(stdout, "\n");
    if (i > 0 &&
	(am_compare((void *)&ams[i], (void *)&ams[i-1]) == 0 ||
	 am_compare((void *)&ams[i-1], (void *)&ams[i]) == 0)) {
      fprintf(stdout, "abmail: overlapping address range, not supported yet.\n");
      ams[i] = ams[i-1];
    }
  }
}

void am_sorts(size_t nam, struct am ams[], int check)
{
  if ((check || nam > USE_BSEARCH_THRESHOLD_N))
    qsort(ams, nam, sizeof(struct am), am_compare);
  if (check)
    am_check(nam, ams);
}

struct am *am_duplicates(size_t nam, struct am ams[])
{
  int i;
  struct am *s;

  if ((s=nam?malloc(sizeof(struct am)*nam):NULL))
    for (i=0; i<nam; i++)
      s[i] = ams[i];
  return s;
}

void amset_append(struct amset *as, struct am am)
{
  if (as->builtin_ams && as->ams == as->builtin_ams)
    as->ams = am_duplicates(as->nam, as->ams);
  as->ams = as->ams?realloc(as->ams, sizeof(struct am)*(as->nam+1)):
    malloc(sizeof(struct am)*(as->nam+1));
  if (as->ams) as->ams[as->nam++] = am;
}

void amset_free(struct amset *as)
{
  if (as->ams != as->builtin_ams) {
    free(as->ams);
    as->nam = sizeof(builtin_ams)/sizeof(builtin_ams[0]);
    as->ams = builtin_ams;
  }
}

void amset_fappend(struct amset *as, FILE *fp, int rc, int info)
{
  char buf[2048];
  int j, n = (!info)?1:2;
  char *p, *q;

  while (fgets(buf, sizeof(buf), fp)) {
    int c = 0, flag = 0;
    struct am am = { "", "", 0, NULL, };

    if (buf[0] == '#' || buf[0] == '\n') continue;
    p = strtok(buf, "\t\n");
    while (p) {
      char cs[] = { '/', '-' };

      switch (c) {
      case 0:
	if (!strchr(p, ':')) {
	  strlcpy(am.a, p, sizeof(am.a));
	  for (j=0; j<sizeof(cs)/sizeof(char); j++)
	    if ((q=strchr(p, cs[j]))) {
	      strlcpy(am.m, (j==0)?q+1:q, sizeof(am.m));
	      if ((q=strchr(am.a, cs[j])))
		*q = '\0';
	      break;
	    }
	  am.rc = rc;
	  flag = !0;
	}
	break;
      case 1:
	am.c = strdup(p);
	break;
      }
      if (++c == n) break;
      p = strtok(NULL, (c+1<n)?"\t\n":"\n");
    }
    if (flag) amset_append(as, am);
  }
}

#if !defined(DISABLE_IPv6)
struct am6 default_am6s_clients[] = {
  { "::1",	/* ::1 */	"128",	-1,	"LOOPBACK"	},
};

struct am6 builtin_am6s[] = {
};

struct am6set am6set_clients = { 0, NULL, NULL }, am6set = {
  sizeof(builtin_am6s)/sizeof(builtin_am6s[0]), builtin_am6s, builtin_am6s
};

#define USE_BSEARCH_THRESHOLD_N 32

struct am6 *am6_bsearch(int n, struct am6 ams[], struct in6_addr a)
{
  struct am6 am_k;

  am_k.ads[0] = a;
  am_k.ads[1] = am6_masks[128];
  am_k.ads[2] = a;
  am_k.ads[3] = a;
  return bsearch(&am_k, ams, n, sizeof(struct am6), am6_compare);
}
struct am6 *am6_lsearch(int n, struct am6 ams[], struct in6_addr a)
{
  int i;

  for (i=0; i<n; i++)
    if (in6_addr_le(ams[i].ads[2], a) && in6_addr_le(a, ams[i].ads[3]))
      break;
  return (i!=n)?&ams[i]:NULL;
}
struct am6 *am6_search(int n, struct am6 ams[], struct in6_addr a)
{
  struct am6 *rv = NULL;

  if (n > USE_BSEARCH_THRESHOLD_N)
    rv = am6_bsearch(n, ams, a);
  else
    rv = am6_lsearch(n, ams, a);
  return rv;
}

void am6_set_spaces(size_t nam, struct am6 ams[])
{
  int i;

  for (i=0; i<nam; i++) {
    if (inet_pton(AF_INET6, ams[i].a, &ams[i].ads[0]) == 0)
      fprintf(stderr, "inet_pton failed for %s\n", ams[i].a);
    am6_set_space(&ams[i]);
  }
}

void am6_check(size_t nam, struct am6 ams[])
{
  int i;
  struct in6_addr one = in6_addr_one();
  char p[64];

  for (i=0; i<nam; i++) {
    if (ams[i].m[0] != '-')
      fprintf(stdout, "  { \"%s\",\t/* %s */\t\"%s\",\t%d,\t\"%s\"\t},",
	      ams[i].a, inet_ntop(AF_INET6, &ams[i].ads[3], p, sizeof(p)), ams[i].m, ams[i].rc, ams[i].c?ams[i].c:"");
    else
      fprintf(stdout, "  { \"%s\",\t\"%s\",\t\t%d,\t\"%s\"\t},",
	      ams[i].a, ams[i].m, ams[i].rc, ams[i].c?ams[i].c:"");
    if (i > 0 &&
	(ams[i-1].rc == ams[i].rc &&
	 in6_addr_eq(in6_addr_add(ams[i-1].ads[3], one), ams[i].ads[2])))
      fprintf(stdout, "\t/* may connect these range */");
    fprintf(stdout, "\n");
    if (i > 0 &&
	(am6_compare((void *)&ams[i], (void *)&ams[i-1]) == 0 ||
	 am6_compare((void *)&ams[i-1], (void *)&ams[i]) == 0)) {
      fprintf(stdout, "abmail: overlapping address range, not supported yet.\n");
      ams[i] = ams[i-1];
    }
  }
}

void am6_sorts(size_t nam, struct am6 ams[], int check)
{
  if ((check || nam > USE_BSEARCH_THRESHOLD_N))
    qsort(ams, nam, sizeof(struct am6), am6_compare);
  if (check)
    am6_check(nam, ams);
}

struct am6 *am6_duplicates(size_t nam, struct am6 ams[])
{
  int i;
  struct am6 *s;

  if ((s=nam?malloc(sizeof(struct am6)*nam):NULL))
    for (i=0; i<nam; i++)
      s[i] = ams[i];
  return s;
}

void am6set_append(struct am6set *as, struct am6 am)
{
  if (as->builtin_ams != NULL && as->ams == as->builtin_ams)
    as->ams = am6_duplicates(as->nam, as->ams);
  as->ams = as->ams?realloc(as->ams, sizeof(struct am6)*(as->nam+1)):
    malloc(sizeof(struct am6)*(as->nam+1));
  if (as->ams) as->ams[as->nam++] = am;
}

void am6set_free(struct am6set *as)
{
  if (as->ams != as->builtin_ams) {
    free(as->ams);
    as->nam = sizeof(builtin_am6s)/sizeof(builtin_am6s[0]);
    as->ams = builtin_am6s;
  }
}

void am6set_fappend(struct am6set *as, FILE *fp, int rc, int info)
{
  char buf[2048];
  int j, n = (!info)?1:2;
  char *p, *q;

  while (fgets(buf, sizeof(buf), fp)) {
    int c = 0, flag = 0;
    struct am6 am = { "", "", 0, NULL, };

    if (buf[0] == '#' || buf[0] == '\n') continue;
    p = strtok(buf, "\t\n");
    while (p) {
      char cs[] = { '/', '-' };

      switch (c) {
      case 0:
	if (strchr(p, ':')) {
	  strlcpy(am.a, p, sizeof(am.a));
	  for (j=0; j<sizeof(cs)/sizeof(char); j++)
	    if ((q=strchr(p, cs[j]))) {
	      strlcpy(am.m, (j==0)?q+1:q, sizeof(am.m));
	      if ((q=strchr(am.a, cs[j])))
		*q = '\0';
	      break;
	    }
	  am.rc = rc;
	  flag = !0;
	}
	break;
      case 1:
	am.c = strdup(p);
	break;
      }
      if (++c == n) break;
      p = strtok(NULL, (c+1<n)?"\t\n":"\n");
    }
    if (flag) am6set_append(as, am);
  }
}
#endif

void amset_read(struct amset *as,
#if !defined(DISABLE_IPv6)
		struct am6set *a6s,
#endif
		int info, int check, int argc, char *argv[])
{
  char *myname = argv?strdup(argv[0]):NULL, *herename = myname?dirname(myname):".";
  char lst_pathname[MAXPATHLEN];
  struct { char *basename; int rc; } lst[] = {
    { "ams_trusted.lst", -1 },
    { "ams_reliable.lst", 0 },
    { "ams_unreliable.lst", 1 },
  };
  int i, flag = 0;
  FILE *fp;

  for (i=0; i<sizeof(lst)/sizeof(lst[0]); i++) {
    sprintf(lst_pathname, "%s/%s", herename, lst[i].basename);
    if (!(fp=fopen(lst_pathname, "r"))) {
      char *home = getenv("HOME");
      sprintf(lst_pathname, "%s/.abmail/%s", home?home:"", lst[i].basename);
      if (!(fp=fopen(lst_pathname, "r"))) {
	sprintf(lst_pathname, PREFIX"/etc/abmail/%s", lst[i].basename);
	if (!(fp=fopen(lst_pathname, "r")))
	  continue;
      }
    }
    amset_fappend(as, fp, lst[i].rc, info);
#if !defined(DISABLE_IPv6)
    rewind(fp);
    am6set_fappend(a6s, fp, lst[i].rc, info);
    fclose(fp);
#endif
    flag = !0;
  }
  if (myname) free(myname);
  am_set_spaces(as->nam, as->ams);
  am_sorts(as->nam, as->ams, check);
#if !defined(DISABLE_IPv6)
  am6_set_spaces(a6s->nam, a6s->ams);
  am6_sorts(a6s->nam, a6s->ams, check);
#endif
}

void amset_clients_read(struct amset *as,
#if !defined(DISABLE_IPv6)
			struct am6set *a6s,
#endif
			int info, int check, int argc, char *argv[])
{
  char *myname = argv?strdup(argv[0]):NULL, *herename = myname?dirname(myname):".";
  char lst_pathname[MAXPATHLEN];
  struct { char *basename; int rc; } lst[] = {
    { "ams_allowable.lst", -1 },
    /*{ "ams_denied.lst", 1 },*/
  };
  int i, flag = 0;
  FILE *fp;

  for (i=0; i<sizeof(lst)/sizeof(lst[0]); i++) {
    sprintf(lst_pathname, "%s/%s", herename, lst[i].basename);
    if (!(fp=fopen(lst_pathname, "r"))) {
      char *home = getenv("HOME");
      sprintf(lst_pathname, "%s/.abmail/%s", home?home:"", lst[i].basename);
      if (!(fp=fopen(lst_pathname, "r"))) {
	sprintf(lst_pathname, PREFIX"/etc/abmail/%s", lst[i].basename);
	if (!(fp=fopen(lst_pathname, "r")))
	  continue;
      }
    }
    amset_fappend(as, fp, lst[i].rc, info);
#if !defined(DISABLE_IPv6)
    rewind(fp);
    am6set_fappend(a6s, fp, lst[i].rc, info);
    fclose(fp);
#endif
    flag = !0;
  }
  if (myname) free(myname);
  if (!flag) {
    as->nam = sizeof(default_ams_clients)/sizeof(struct am);
    as->ams = as->builtin_ams = default_ams_clients;
    am_set_spaces(as->nam, as->ams);
#if !defined(DISABLE_IPv6)
    a6s->nam = sizeof(default_am6s_clients)/sizeof(struct am6);
    a6s->ams = a6s->builtin_ams = default_am6s_clients;
    am6_set_spaces(a6s->nam, a6s->ams);
#endif
    {
      char s[2048];
      struct hostent *he;
      int i;

      gethostname(s, sizeof(s));
      if ((he=gethostbyname(s)))
	for (i=0; he->h_addr_list[i]; i++) {
	  if (!am_lsearch(as->nam, as->ams, *(struct in_addr *)he->h_addr_list[i])) {
	    struct am am = { "", "", -1, NULL, };

	    strlcpy(am.a, inet_ntoa((*(struct in_addr *)he->h_addr_list[i])), sizeof(am.a));
	    amset_append(as, am);
#ifdef DEBUG
	    fprintf(stderr, "server name:\t%s\nserver addr:\t%s\n", s, am.a);
#endif
	  }
	}
#if !defined(DISABLE_IPv6)
#if !defined(USE_GETHOSTBYNAME2)
      {
	struct addrinfo hints, *res, *r;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	if (!getaddrinfo(s, NULL, &hints, &res))
	  for (r=res; r; r=r->ai_next)
	    if (r->ai_family == AF_INET6 &&
		!am6_lsearch(a6s->nam, a6s->ams, ((struct sockaddr_in6 *)r->ai_addr)->sin6_addr)) {
	      struct am6 am = { "", "", -1, NULL, };

	      inet_ntop(AF_INET6, &((struct sockaddr_in6 *)r->ai_addr)->sin6_addr, am.a, sizeof(am.a));
	      am6set_append(a6s, am);
#ifdef DEBUG
	      fprintf(stderr, "server name:\t%s\nserver addr:\t%s\n", s, am.a);
#endif
	    }
      }
#else
      if ((he=gethostbyname2(s, AF_INET6)))
	for (i=0; he->h_addr_list[i]; i++) {
	  if (!am6_lsearch(a6s->nam, a6s->ams, *(struct in6_addr *)he->h_addr_list[i])) {
	    struct am6 am = { "", "", -1, NULL, };

	    inet_ntop(AF_INET6, he->h_addr_list[i], am.a, sizeof(am.a));
	    am6set_append(a6s, am);
#ifdef DEBUG
	    fprintf(stderr, "server name:\t%s\nserver addr:\t%s\n", s, am.a);
#endif
	  }
	}
#endif
#endif
    }
  }
  am_set_spaces(as->nam, as->ams);
  am_sorts(as->nam, as->ams, check);
#if !defined(DISABLE_IPv6)
  am6_set_spaces(a6s->nam, a6s->ams);
  am6_sorts(a6s->nam, a6s->ams, check);
#endif
}
