/*
 * abaddr.x - RPC protocol of abaddrd client and server
 *
 * Copyright (C) 2008 Taiji Yamada <taiji@aihara.co.jp>
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Affero General Public License for more details.
 */
#if defined(RPC_SVC)
%#define svc_run abaddr_svc_run
%#define main abaddr_main
#endif

enum abaddrf { /* http://www.iana.org/assignments/address-family-numbers */
  ABADDRF_INET	= 1,
  ABADDRF_INET6	= 2
};

union abaddrs switch (enum abaddrf af) {
  case ABADDRF_INET:
    unsigned char s[8];		/* twice of struct in_addr */
  case ABADDRF_INET6:
    unsigned char s6[32];	/* twice of struct in6_addr */
};

struct abaddrattr {
  abaddrs addr;
  int rc;	/* related code */
  long lt;	/* life time [sec], 0: infinite, -1: immediate, -2: conditional */
  string c<128>;	/* characters */
};

enum abaddrproc_stat {
  ABADDRPROC_OK,
  ABADDRPROC_EACCES,
  ABADDRPROC_EIO,
  ABADDRPROC_ENOTSUP,
  ABADDRPROC_NG
};

enum abaddr_req {
  ABADDR_REQ_DB_LIFE,
  ABADDR_REQ_DB_TRNC,
  ABADDR_REQ_DB_LOAD,
  ABADDR_REQ_DB_SAVE
};

struct abaddr_get_stat {
  abaddrproc_stat stat;
  abaddrattr at;
};

program ABADDRPROG {
  version ABADDRVERS {
    abaddrproc_stat ABADDRPROC_REQ_SERV(abaddr_req) = 1;
    abaddrproc_stat ABADDRPROC_ADDR_SET(abaddrattr) = 2;
    abaddr_get_stat ABADDRPROC_ADDR_GET(abaddrattr) = 3;
  } = 1;
} = 900000000;
