/*
 * abcidr.h - utilities for CIDR in abmail
 *
 * Copyright (C) 2007-2008 Taiji Yamada <taiji@aihara.co.jp>
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Affero General Public License for more details.
 */
#ifndef _abcidr_h_
#define _abcidr_h_
#include <stdio.h>	/* FILE, etc */
#include <netinet/in.h>	/* in_addr, etc */

struct am {
  char a[16], m[1+16];
  int rc;
  char *c;
  struct in_addr ads[4];	/* 0: address, 1: mask, 2: network, 3: broadcast */
};

int am_compare(const void *a, const void *b);
extern struct in_addr am_masks[32+1];
void am_init_masks(void);
int am_set_space(struct am *am);
int am_put_cidr(struct am *am, int lm);
int am_put_mask(struct am *am);
int am_put_range(struct am *am);
int am_fprint_range(struct am *am, FILE *fp);
int am_fprint_cidrs(struct am *am, FILE *fp);

#if !defined(DISABLE_IPv6)
struct in6_addr in6_addr_one(void);
int in6_addr_cmp(struct in6_addr a, struct in6_addr b);
struct in6_addr in6_addr_add(struct in6_addr a, struct in6_addr b);
struct in6_addr in6_addr_sub(struct in6_addr a, struct in6_addr b);
#define in6_addr_eq(a, b) (in6_addr_cmp(a, b) == 0)
#define in6_addr_ne(a, b) (in6_addr_cmp(a, b) != 0)
#define in6_addr_lt(a, b) (in6_addr_cmp(a, b) < 0)
#define in6_addr_gt(a, b) (in6_addr_cmp(a, b) > 0)
#define in6_addr_le(a, b) (in6_addr_cmp(a, b) <= 0)
#define in6_addr_ge(a, b) (in6_addr_cmp(a, b) >= 0)

struct am6 {
  char a[64], m[1+64];
  int rc;
  char *c;
  struct in6_addr ads[4];	/* 0: address, 1: mask, 2: network, 3: netend */
};

int am6_compare(const void *a, const void *b);
extern struct in6_addr am6_masks[128+1];
void am6_init_masks(void);
int am6_set_space(struct am6 *am);
int am6_put_cidr(struct am6 *am, int lm);
int am6_put_mask(struct am6 *am);
int am6_put_range(struct am6 *am);
int am6_fprint_range(struct am6 *am, FILE *fp);
int am6_fprint_cidrs(struct am6 *am, FILE *fp);
#endif
#endif
