/*
 * abmailc.c - (ab)mail converter
 *
 * Copyright (C) 2007-2008 Taiji Yamada <taiji@aihara.co.jp>
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Affero General Public License for more details.
 */
/*
 * usage:
 *	$ ./abmailc [-E|[-b-] [-D-] [-S] [-B-] [-s] [-w width] [-c length] [-l nlines]] [- mail|mhdir|maildir [...]] | lv -Iu8
 */
#include <stdio.h>	/* fopen, etc */
#include <string.h>	/* strcmp, etc */
#include <ctype.h>	/* isspace, etc */
#include <libgen.h>	/* dirname, etc */
#include <sys/types.h>	/* sys/stat.h, dirent.h, etc */
#include <sys/stat.h>	/* stat, etc */
#include <dirent.h>	/* opendir, etc */
#include <pcre.h>	/* pcre, etc */
#include <sys/param.h>	/* MAXPATHLEN, etc */
#include <errno.h>	/* errno */
#include <iconv.h>	/* iconv */

//#define DEBUG_I18N

#define min(a, b) (((a)<(b))?(a):(b))
#define max(a, b) (((a)>(b))?(a):(b))
#define lbu(l, b, u) min(max((l), (b)), (u))

#define min(a, b) (((a)<(b))?(a):(b))
#define max(a, b) (((a)>(b))?(a):(b))
#define lbu(l, b, u) min(max((l), (b)), (u))

#if !defined(HAVE_STRLCPY)
#define strlcpy(d, s, z) do { strncpy(d, s, z); if ((z) > 0) (d)[(z)-1] = '\0'; } while (0)
#endif

#if !defined(HAVE_STRLCAT)
#define strlcat(d, s, z) do { if ((z) > strlen(d)) strncat(d, s, (z)-strlen(d)-1); } while (0)
#endif

#if !defined(PREFIX)
#define PREFIX "/opt/local"
#endif

int aredigit(const char *s)
{
  size_t i, z;

  if (!s) return 0;
  if (!(z = strlen(s)))
    return 0;
  for (i=0; i<z; i++)
    if (!isdigit(s[i]))
      return 0;
  return !0;
}

int arespace(const char *s)
{
  size_t i, z;

  if (!s) return 0;
  if (!(z = strlen(s)))
    return 0;
  for (i=0; i<z; i++)
    if (!isspace(s[i]))
      return 0;
  return !0;
}

char *strnspace(const char *s)
{
  char *r;

  if (!s) return NULL;
  for (r=(char *)s; *r!='\0'; r++)
    if (!isspace(*r))
      break;
  return (*r == '\0') ? NULL : r;
}

char *strspace(const char *s)
{
  char *r;

  if (!s) return NULL;
  for (r=(char *)s; *r!='\0'; r++)
    if (isspace(*r))
      break;
  return (*r == '\0') ? NULL : r;
}

size_t strchop(char *s, const char *sep)
{
  size_t i, j, z = strlen(s), l = strlen(sep);

  i = z;
  if (z)
    do {
      i--;
      for (j=0; j<l; j++)
	if (s[i] == sep[j]) {
	  s[i] = '\0';
	  break;
	}
      if (j == l) {
	i++;
	break;
      }
    } while (i != 0);
  return i;
}

char *strchrs(const char *s, const char *sep)
{
  size_t i, j, z = strlen(s), l = strlen(sep);

  for (i=0; i<z; i++)
    for (j=0; j<l; j++)
      if (s[i] == sep[j])
	return (char *)&s[i];
  return NULL;
}

char *strnchrs(const char *s, const char *sep, long n, long *np)
{
  size_t i, j, z = strlen(s), l = strlen(sep), c = 0;

  if (np) *np = n;
  for (i=0; i<z; i++)
    for (j=0; j<l; j++)
      if (s[i] == sep[j] && ++c == n)
	return (char *)&s[i];
  if (np) *np = c;
  return NULL;
}

int strlenc_cc(char *dst, const char *src, size_t dz, const char *dc, const char *sc)
{
  int rc = 0, en = 0;
  iconv_t cd = (iconv_t)-1;
  size_t r = (size_t)-1;
  const char *ip = src;
  char *op = dst;
  size_t il = strlen(src), ol = dz - 1;

  if ((cd=iconv_open(dc, sc)) != (iconv_t)-1) {
    if ((r = iconv(cd, &ip, &il, &op, &ol)) == (size_t)-1)
      en = errno;
    else
      r = iconv(cd, NULL, NULL, &op, &ol);
    *op = '\0';
    iconv_close(cd);
  }
  if (cd != (iconv_t)-1 && r != (size_t)-1)
    rc = !0;
  return rc;
}

int fputs_cc(const char *s, FILE *fp, const char *dc, const char *sc)
{
  if (dc == sc)
    return fputs(s, fp);
  else {
    char b[32768] = "";

    strlenc_cc(b, s, sizeof(b), dc, sc);
    return fputs(b, fp);
  }
}

int arespace_wc(const char *src, const char *sc)
{
  const char *dc = "ucs-4be";
  char dst[4];
  size_t dz = sizeof(dst);
  int rc = 0, en = 0;
  iconv_t cd = (iconv_t)-1;
  size_t r = (size_t)-1;
  const char *ip = src;
  char *op = dst;
  size_t il = strlen(src), ol = dz;

  if (il && (cd=iconv_open(dc, sc)) != (iconv_t)-1) {
    rc = !0;
    do {
      en = (r = iconv(cd, &ip, &il, &op, &ol)) == (size_t)-1 ?
	errno : 0;
      if (!(( memcmp(&dst[0], "\x00\x00\x00",	3) == 0	&&
	     (isspace(dst[3]) || dst[3] == '\xa0'))	||
	    ( memcmp(&dst[0], "\x00\x00\x20",	3) == 0	&&
	     (('\x00' <= dst[3] && dst[3] <= '\x0b') ||
	      dst[3] == '\x2f' || dst[3] == '\x5f'))	||
	    ( memcmp(&dst[0], "\x00\x00",	2) == 0	&&
	     (memcmp(&dst[2], "\x30\x00",	2) == 0	||
	      memcmp(&dst[2], "\xfe\xff",	2) == 0)))) {
	rc = 0;
	break;
      }
      op = dst;
      ol = dz;
    } while (en == E2BIG);
    iconv_close(cd);
  }
  return rc;
}

char *strn_wc(const char *src, const char *sc, long n, long *np)
{
  const char *dc = "ucs-4be";
  char dst[4], *p = NULL;
  size_t dz = sizeof(dst), c = 0;
  int en = 0;
  iconv_t cd = (iconv_t)-1;
  size_t r = (size_t)-1;
  const char *ip = src;
  char *op = dst;
  size_t il = strlen(src), ol = dz;

  if (np) *np = n;
  if (il && (cd=iconv_open(dc, sc)) != (iconv_t)-1) {
    do {
      en = (r = iconv(cd, &ip, &il, &op, &ol)) == (size_t)-1 ?
	errno : 0;
      if (++c == n) {
	p = (char *)ip;
	break;
      }
      op = dst;
      ol = dz;
    } while (en == E2BIG);
    iconv_close(cd);
  }
  if (np) *np = c;
  return p;
}

void strncpy_nonl(char *d, const char *s, size_t z)
{
  size_t i, j;

  for (j=0, i=0; i<z; i++)
    if (s[i] == '\n' || s[i] == '\r')
      while (i+1<z && isspace(s[i+1])) i++;
    else {
      d[j++] = s[i];
      if (s[i] == '\0') break;
    }
}
#define strlcpy_nonl(d, s, z) do { strncpy_nonl(d, s, z); if ((z) > 0) d[(z)-1] = '\0'; } while (0)

const char strlenc_b_tbl[] =
  "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

void strlenc_b(char *d, const char *src, size_t dz, size_t sz)
{
  size_t i, j;
  unsigned char *s = (unsigned char *)src;

  for (j=0, i=0; i<sz && j<dz-1; i+=3) {
    d[j++] = strlenc_b_tbl[ (s[i+0]&0xfc)>>2                        ];
    if (j>=dz-1) break;
    d[j++] = strlenc_b_tbl[((s[i+0]&0x03)<<4) | ((s[i+1] & 0xf0)>>4)];
    if (j>=dz-1) break;
    d[j++] = strlenc_b_tbl[((s[i+1]&0x0f)<<2) | ((s[i+2] & 0xc0)>>6)];
    if (j>=dz-1) break;
    d[j++] = strlenc_b_tbl[  s[i+2]&0x3f                            ];
  }
  d[j] = '\0';
}

const char strldec_b_tbl[] =
  "\x3e\x00\x00\x00\x3f\x34\x35\x36\x37\x38\x39\x3a\x3b\x3c\x3d\x00"
  "\x00\x00\x00\x00\x00\x00\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09"
  "\x0a\x0b\x0c\x0d\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19"
  "\x00\x00\x00\x00\x00\x00\x1a\x1b\x1c\x1d\x1e\x1f\x20\x21\x22\x23"
  "\x24\x25\x26\x27\x28\x29\x2a\x2b\x2c\x2d\x2e\x2f\x30\x31\x32\x33";

void strldec_b(char *d, const char *src, size_t dz, size_t sz)
{
  size_t i, j;

  for (j=0, i=0; i<sz && j<dz-1; i+=4) {
    unsigned char s[] = {
      src[i]-43,
      (i+1<sz ? src[i+1] : '=')-43,
      (i+2<sz ? src[i+2] : '=')-43,
      (i+3<sz ? src[i+3] : '=')-43
    };

    if (!(s[0] < sizeof(strldec_b_tbl)-1 &&
	  s[1] < sizeof(strldec_b_tbl)-1 &&
	  s[2] < sizeof(strldec_b_tbl)-1 &&
	  s[3] < sizeof(strldec_b_tbl)-1))
      break;
    d[j++] = ( strldec_b_tbl[s[0]]      <<2) | ((strldec_b_tbl[s[1]]&0x30)>>4);
    if (j>=dz-1) break;
    d[j++] = ((strldec_b_tbl[s[1]]&0x0f)<<4) | ((strldec_b_tbl[s[2]]&0x3c)>>2);
    if (j>=dz-1) break;
    d[j++] = ((strldec_b_tbl[s[2]]&0x03)<<6) | ( strldec_b_tbl[s[3]]         );
  }
  d[j] = '\0';
}

void strldec_q(char *d, const char *s, size_t dz, const size_t sz)
{
  size_t i, j;

  for (j=0, i=0; i<sz && j<dz-1; i++) {
    if (s[i] == '=' && isxdigit(s[i+1]) && isxdigit(s[i+2])) {
      char b[3];
      int x;

      strlcpy(b, &s[i+1], sizeof(b));
      sscanf(b, "%x", &x);
      d[j++] = x; if (j>=dz-1) break;
      i += 2;
    }
    else {
      d[j++] = s[i]; if (j>=dz-1) break;
    }
  }
  d[j] = '\0';
}

enum {
  EC_none,
  EC_b,
  EC_q,
  EC_unknown,
};
const char *encodes[EC_unknown] = {
  NULL,
  "?b?",
  "?q?",
};
const char *encodings[EC_unknown] = {
  NULL,
  "base64",
  "quoted-printable",
};

void strldec_ec(char *d, const char *s, size_t dz, const size_t sz, int k)
{
  switch (k) {
  case EC_b:
    strldec_b(d, s, dz, sz);
    break;
  case EC_q:
    strldec_q(d, s, dz, sz);
    break;
  }
}

enum {
  CC_none,
  CC_utf_8,
  CC_iso_2022_jp,
  CC_shift_jis,
  CC_sjis,
  CC_gb2312,
  CC_big5,
  CC_euc_kr,
  CC_euc_jp,
  CC_iso_8859_1,
  CC_iso_8859_2,
  CC_iso_8859_4,
  CC_iso_8859_9,
  CC_iso_8859_15,
  CC_iso_8859_16,
  CC_koi8_r,
  CC_windows_1250,
  CC_windows_1251,
  CC_windows_1252,
  CC_windows_1254,
  CC_windows_1255,
  CC_windows_1256,
  CC_windows_1257,
  CC_us_ascii,
  CC_utf_7,
  CC_ks_c_5601_1987,
  CC_unicode_1_1_utf_7,
  CC_cp932,
  CC_cp936,
  CC_cp949,
  CC_cp950,
  CC_gbk,
  CC_gb18030,
  CC_iso_2022_jp_3,
  CC_unknown
};
const char *charcodes[CC_unknown] = {
  NULL,
  "utf-8",
  "iso-2022-jp",
  "shift_jis",
  "shift-jis",
  "gb2312",
  "big5",
  "euc-kr",
  "euc-jp",
  "iso-8859-1",
  "iso-8859-2",
  "iso-8859-4",
  "iso-8859-9",
  "iso-8859-15",
  "iso-8859-16",
  "koi8-r",
  "windows-1250",
  "windows-1251",
  "windows-1252",
  "windows-1254",
  "windows-1255",
  "windows-1256",
  "windows-1257",
  "us-ascii",
  "utf-7",
  "ks_c_5601-1987",
  "unicode-1-1-utf-7",
  "cp932",
  "cp936",
  "cp949",
  "cp950",
  "gbk",
  "gb18030",
  "iso-2022-jp-3",
};

int strldec_cc(char *dst, const char *src, size_t dz, const char *dc, const char *sc)
{
#define uc(c) ((unsigned char)c)
#define is_cp5022x(p)						\
  ((*(p) == '\x2d' ||						\
    ('\x79' <= *(p) && *(p) <= '\x7c') ||			\
    (uc('\x7f') <= uc(*(p)) && uc(*(p)) <= uc('\x92'))) &&	\
   '\x21' <= *((p)+1) && *((p)+1) <= '\x7e')
  int rc = 0, en = 0;
  iconv_t cd = (iconv_t)-1;
  size_t r = (size_t)-1;
  const char *ip = src;
  char *op = dst;
  size_t il = strlen(src), ol = dz - 1;

  if ((cd=iconv_open(dc, sc)) != (iconv_t)-1) {
    if ((r = iconv(cd, &ip, &il, &op, &ol)) == (size_t)-1)
      en = errno;
    else
      r = iconv(cd, NULL, NULL, &op, &ol);
    if (en == EILSEQ) {
      char *q;

      if (strncasecmp(sc, "iso-2022-jp", sizeof("iso-2022-jp")-1) == 0 &&
	  (q=strstr(src, "\x1b$B")) && q <= ip) {	/* for cp5022x */
	const char *p;

	while (en == EILSEQ &&
	       (is_cp5022x((p=ip)) ||
		(strncmp(ip, "\x1b$B", sizeof("\x1b$B")-1) == 0 &&
		 is_cp5022x((p=ip+sizeof("\x1b$B")-1))))) {
	  const char b[] = { *p, *(p+1) };

	  strncpy((char *)p, "\".", sizeof("\".")-1);	/* ! */
	  if ((r = iconv(cd, &ip, &il, &op, &ol)) == (size_t)-1)
	    en = errno;
	  strncpy((char *)p, b, sizeof(b));
	}
	if (r != (size_t)-1)
	  r = iconv(cd, NULL, NULL, &op, &ol);
      }
    }
    *op = '\0';
    iconv_close(cd);
  }
  if (cd != (iconv_t)-1 && r != (size_t)-1)
    rc = !0;
#if defined(DEBUG_I18N)
  fprintf(stdout, "Debug: %s <%c %s\n", dc, rc?'=':'-', (!sc||sc[0]=='\0')?"unknown":sc);
#endif
  return rc;
#undef uc
#undef is_cp5022x
}

enum {
  WC_none,
  WC_iso_2022_jp_3,
  WC_euc_kr,
  WC_cp932,
  WC_cp936,
  WC_euc_jp,
  WC_big5,
  WC_windows_1251,
  WC_iso_8859_15,
  WC_unknown
};
const int wccodes[WC_unknown] = {
  CC_none,
  CC_iso_2022_jp_3,
  CC_euc_kr,
  CC_cp932,
  CC_cp936,
  CC_euc_jp,
  CC_big5,
  CC_windows_1251,
  CC_iso_8859_15,
};
int strldec_wc(char *dst, const char *src, size_t dz, const char *dc, const char *sc)
{
  int wc = CC_none;
  size_t i;

  for (i=0; i<strlen(src); i++)
    if (src[i] == '\x1b') {
      wc = CC_iso_2022_jp;
      break;
    }
    else if (src[i] < 0) {
      wc = CC_unknown;
      break;
    }
  if (!wc && !sc)
    return wc;

  if (!sc) {
    for (i=1; i<sizeof(wccodes)/sizeof(int); i++)
      if (strldec_cc(dst, src, dz, dc, charcodes[wccodes[i]]))
	break;
    if (i == sizeof(wccodes)/sizeof(int))
      strlcpy(dst, src, dz);
    wc = wccodes[i];
  }
  else if ((strcasecmp(sc, "iso-2022-jp") == 0 && wc == CC_unknown) ||
	   strcasecmp(sc, "gb2312") == 0) {
    if (strldec_cc(dst, src, dz, dc, charcodes[CC_cp932]))
      wc = CC_cp932;
    else if (strldec_cc(dst, src, dz, dc, charcodes[CC_cp936]))
      wc = CC_cp936;
    else
      wc = strldec_wc(dst, src, dz, dc, charcodes[CC_none]);
  }
  else if (strcasecmp(sc, "ks_c_5601-1987") == 0) {
    if (strldec_cc(dst, src, dz, dc, charcodes[CC_cp949]))
      wc = CC_cp949;
    else
      wc = strldec_wc(dst, src, dz, dc, charcodes[CC_none]);
  }
  else if (strcasecmp(sc, "big5") == 0) {
    if (strldec_cc(dst, src, dz, dc, charcodes[CC_cp950]))
      wc = CC_cp950;
    else
      wc = strldec_wc(dst, src, dz, dc, charcodes[CC_none]);
  }
  else
    wc = strldec_wc(dst, src, dz, dc, charcodes[CC_none]);
  return wc;
}

int strldec_mm(char *dst, const char *src, size_t dz, const char *dc)
{
  int cc = CC_none, subst = 0;
  size_t j, j0, k, s;
  int i;
  char *p, *q = NULL;

  strlcpy(dst, "", dz);
  for (j0=0, j=0; j<(s=strlen(src)); j++) {
    if (!(p = strstr(&src[j], "=?"))) break;
    j = p - src;
    for (i=1; i<sizeof(charcodes)/sizeof(char *); i++) {
      for (k=1; k<sizeof(encodes)/sizeof(char *); k++) {
	char ee[256] = "=?";
	size_t l;

	strlcat(ee, charcodes[i], sizeof(ee));
	strlcat(ee, encodes[k], sizeof(ee));
	l = strlen(ee);
	if (strncasecmp(&src[j], ee, l) == 0 && (p = strstr(&src[j+l], "?="))) {
	  char b[j-j0+1], s[p-&src[j+l]+1], d[1024];

	  while (q && isspace(*q)) { j0++; q++; }
	  strncpy(b, &src[j0], j-j0); b[j-j0] = '\0';
	  /*strncpy_nonl(b, &src[j0], j-j0); b[j-j0] = '\0';*/ /* ! */
	  strlcat(dst, b, dz);
	  strlcpy(s, &src[j+l], sizeof(s));
	  strldec_ec(d, s, sizeof(d), strlen(s)+1, k);
	  {
	    char e[32768];
	    int yacc;

	    if (strldec_cc(e, d, sizeof(e), dc, charcodes[i]))
	      cc = i;
	    else
	      if ((yacc=strldec_wc(e, d, sizeof(e), dc, charcodes[i]))) {
		cc = yacc;
		subst = !0;
	      }
	    if (cc) {
#if defined(DEBUG_I18N)
	      strlcat(dst, subst?"{{":"[[", dz);
	      strlcat(dst, charcodes[cc], dz);
	      if (subst) {
		strlcat(dst, "!", dz);
		strlcat(dst, charcodes[i], dz);
	      }
	      strlcat(dst, ":", dz);
	      strlcat(dst, e, dz);
	      strlcat(dst, !subst?"]]":"}}", dz);
#else
	      strlcat(dst, e, dz);
#endif
	    }
	    else {
	      char b[p+sizeof("?=")-1-&src[j]+1];

	      strncpy(b, &src[j], p+sizeof("?=")-1-&src[j]); b[p+sizeof("?=")-1-&src[j]] = '\0';
	      strlcat(dst, "{{unknown:", dz);
	      strlcat(dst, b, dz);
	      strlcat(dst, "}}", dz);
	    }
	  }
	  q = p + sizeof("?=")-1;
	  j += l + (q-&src[j+l]) - 1;
	  j0 = j + 1;
	  i = sizeof(charcodes)/sizeof(char *);
	  break;
	}
      }
    }
  }
  if (j0 < s)
    strlcat(dst, &src[j0], dz);
  return cc;
}

enum {
  CT_none,
  //CT_text_html,
  CT_text_plain,
  CT_unknown,
};
const char *content_types[CT_unknown] = {
  NULL,
  //"text/html",
  "text/plain",
};

#define OE_File			(1L<<1)
#define OE_Envelope_From	(1L<<2)
#define OE_Return_Path		(1L<<3)
#define OE_Received		(1L<<4)
#define OE_To			(1L<<5)
#define OE_From			(1L<<6)
#define OE_Subject		(1L<<7)
#define OE_Message_Id		(1L<<8)
#define OE_eoh			(1L<<9)

static char *items[] = {
  "Received:",
  "To:",
  "From:",
  "Subject:",
  "Message-Id:",
};
enum {
  HI_Received,
  HI_To,
  HI_From,
  HI_Subject,
  HI_Message_Id,
  HI_unknown,
};

struct abmail {
  int hda, iss, ibl;	/* hidden dot algorithm, ignore spaces, ignore blank line */
  int ce, ceo;		/* character encoding to decode, character encoding to output */
  int body;		/* process body */
  int lom, loh, hop;	/* level of multipart, level of header, header of part */
  char from[3][2048];	/* 0: full address, 1: logname, 2: domain */
  char retp[3][2048];	/* 0: full address, 1: logname, 2: domain */
  char item[HI_unknown][8192];	/* 0: received, 1: recipient, 2: sender, 3: subject, 4: message-id */
  char info[3][1024];	/* 0: content-type, 1: charset, 2: content-transfer-encoding */
  char boundary[5][2048];
  char buf[32768], line[32768];
  int ct, cs, cte, boe;	/* content-type, charset, content-transfer-encoding, begin of encoding */
  char *csbuf, *ctebuf;
  struct output {
    long em;		/* entry mask to output */
    int dot;		/* message separator by hidden dot algorithm */
    long lh;		/* the limit length of header strings */
    int lc;		/* last char */
    char ls[32768];	/* last string for white spaces */
    long lz;		/* the length of last string */
    long nb, lb;	/* the length of content, the limit length of content */
    long nl, ll;	/* the number of lines, the limit number of lines */
  } op;
  char filename[2048];
} me = {
  !0, 0, !0,
  CC_utf_7, CC_utf_8,
  !0,
  0, 0, 0,
  { "", "", "", }, { "", "", "", }, { "", "", "", "" }, { "", "", "" }, { "","","","","" },
  "", "",
  0, 0, 0, 0,
  NULL, NULL,
  { ~0L, '.', 0, 0, "", 0, 0, 0, 0, 0 },
  "",
};

struct pz {
  char *p;
  size_t z;
};

int pcre_set_substring_pz_array(char *subject, int *ovector,
				int stringcount, struct pz a[])
{
  int i;

  for (i=0; i<stringcount; i++) {
    a[i].p = subject + ovector[i*2];
    a[i].z = ovector[i*2+1] - ovector[i*2];
  }
  return 0;
}

int amp_from(struct abmail *m, int n, const char **l, struct pz a[]);
int amp_retp(struct abmail *m, int n, const char **l, struct pz a[]);
int amp_item(struct abmail *m, int n, const char **l, struct pz a[]);
int amp_local(struct abmail *m, int n, const char **l, struct pz a[]);
int amp_mail(struct abmail *m, int n, const char **l, struct pz a[]);
int amp_mailenc(struct abmail *m, int n, const char **l, struct pz a[]);
int amp_boundary(struct abmail *m, int n, const char **l, struct pz a[]);
int amp_bomail(struct abmail *m, int n, const char **l, struct pz a[]);

#define RE_BYM "\\s+by\\s+.*"

struct headre {
  int sw;	/* switch on: !0, off: 0, loh on: >1  */
  int n;	/* nth header matching, n=-1 means for all the same headers */
  int neg;	/* negative matching on: !0, off: 0 */
  char *header;	/* header string */
  char *re;	/* regular expression */
  int (*proc)(struct abmail *, int, const char **, struct pz *);	/* process at matching */
  char *msg;	/* message string */
  int nv;	/* the number of vectors */
} headres[] = {
  { !0,  0,  0, "From ", "\\s*(([^\\s]*)@([^@\\s]*)|[^@\\s]*)", amp_from, NULL },
  { !0,  0,  0, "Return-Path:", "\\s*(<([^\\s]*)@([^@\\s]*)|[^@\\s]*>)", amp_retp, NULL },
  { !0, -1,  0, "Received:", "(?:"RE_BYM"(?:\\s*.*)*|%s)\\s*$", amp_local, NULL },
  { !0, -1,  0, "", "(Received:|To:|From:|Subject:|Message-Id:)\\s*([\\S\\s]*)", amp_item, NULL },
  { !0,  0, !0, "Content-Type:", "\\s*([^\\s;]*)(?:[;\\s]*.*;\\s*charset\\s*=\\s*\"?([^\\s\";]*))?", amp_mail, NULL },
  { !0,  0,  0, "Content-Transfer-Encoding:", "\\s*(base64|quoted-printable)", amp_mailenc, NULL },
  { !0,  0, !0, "Content-Type:", "\\s*multipart/(?:report|alternative|related|mixed);\\s*.*\\s*boundary=\\s*\"?([^\\s\";]*)", amp_boundary, NULL },
  {  2,  0,  0, "Content-Type:", "\\s*(message/rfc822|text/rfc822-headers)", amp_bomail, NULL },
  {  2,  0, !0, "Content-Type:", "\\s*([^\\s;]*)(?:[;\\s]*.*;\\s*charset\\s*=\\s*\"?([^\\s\";]*))?", amp_mail, NULL },
  {  2,  0,  0, "Content-Transfer-Encoding:", "\\s*(base64|quoted-printable)", amp_mailenc, NULL },
  {  2,  0, !0, "Content-Type:", "\\s*multipart/(?:alternative|related|mixed);\\s*.*\\s*boundary=\\s*\"?([^\\s\";]*)", amp_boundary, NULL },
  {  3,  0,  0, "From ", "\\s*(([^\\s]*)@([^@\\s]*)|[^@\\s]*)", amp_from, NULL },
  {  3,  0,  0, "Return-Path:", "\\s*(<([^\\s]*)@([^@\\s]*)|[^@\\s]*>)", amp_retp, NULL },
  {  3, -1,  0, "Received:", "(?:"RE_BYM"(?:\\s*.*)*|%s)\\s*$", amp_local, NULL },
  {  3, -1,  0, "", "(Received:|To:|From:|Subject:|Message-Id:)\\s*([\\S\\s]*)", amp_item, NULL },
  {  3,  0, !0, "Content-Type:", "\\s*([^\\s;]*)(?:[;\\s]*.*;\\s*charset\\s*=\\s*\"?([^\\s\";]*))?", amp_mail, NULL },
  {  3,  0,  0, "Content-Transfer-Encoding:", "\\s*(base64|quoted-printable)", amp_mailenc, NULL },
  {  3,  0, !0, "Content-Type:", "\\s*multipart/(?:alternative|related|mixed);\\s*.*\\s*boundary=\\s*\"?([^\\s\";]*)", amp_boundary, NULL },
};
enum {
  HR_from,
  HR_retp,
  HR_local,
  HR_item,
  HR_mail,
  HR_mailenc,
  HR_boundary,
  HR_bomail,
  HR_mail2,
  HR_mailenc2,
  HR_boundary2,
  HR_bofrom,
  HR_boretp,
  HR_bolocal,
  HR_boitem,
  HR_mail3,
  HR_mailenc3,
  HR_boundary3,
};

int amp_from(struct abmail *m, int n, const char **l, struct pz a[])
{
  int rc = 0, i;

  if (m->from[0][0] == '\0')
    for (i=0; i<3; i++)
      if (n > i+1) {
	char b[sizeof(m->from[i])];

	if (strldec_wc(b, l[i+1], sizeof(m->from[i]), charcodes[m->ce], charcodes[CC_none]))
	  strlcpy_nonl(m->from[i], b, sizeof(m->from[i]));
	else
	  strlcpy_nonl(m->from[i], l[i+1], sizeof(m->from[i]));
      }
  return rc;
}

int amp_retp(struct abmail *m, int n, const char **l, struct pz a[])
{
  int rc = 0, i;

  if (m->retp[0][0] == '\0')
    for (i=0; i<3; i++)
      if (n > i+1) {
	char b[sizeof(m->retp[i])];

	if (strldec_wc(b, l[i+1], sizeof(m->retp[i]), charcodes[m->ce], charcodes[CC_none]))
	  strlcpy_nonl(m->retp[i], b, sizeof(m->retp[i]));
	else
	  strlcpy_nonl(m->retp[i], l[i+1], sizeof(m->retp[i]));
      }
  return rc;
}

int amp_item(struct abmail *m, int n, const char **l, struct pz a[])
{
  int rc = 0;

  if (n > 2) {
    int i;
    char buf[2][sizeof(m->item[0])];
    int cc, wc;

    for (i=0; i<sizeof(items)/sizeof(char *); i++)
      if (strcasecmp(l[1], items[i]) == 0)
	break;
    if (m->item[i][0] != '\0')
      return rc;
    switch (i) {
    case HI_Received:
    case HI_Message_Id:
      cc = strldec_cc(buf[0], l[2], sizeof(m->item[0]), charcodes[m->ce], charcodes[CC_us_ascii]);
      if ((wc = strldec_wc(buf[1], buf[0], sizeof(m->item[0]), charcodes[m->ce], charcodes[CC_none])))
	strlcpy(m->item[i], buf[1], sizeof(m->item[0]));
      else
	strlcpy(m->item[i], buf[0], sizeof(m->item[0]));
      strchop(m->item[i], "\t\n\v\f\r ");
      break;
    case HI_To:
    case HI_From:
    case HI_Subject:
      cc = strldec_mm(buf[0], l[2], sizeof(m->item[0]), charcodes[m->ce]);
      if ((wc = strldec_wc(buf[1], buf[0], sizeof(m->item[0]), charcodes[m->ce], charcodes[CC_none])))
	strlcpy_nonl(m->item[i], buf[1], sizeof(m->item[0]));
      else
	strlcpy_nonl(m->item[i], buf[0], sizeof(m->item[0]));
      break;
    }
  }
  return rc;
}

int amp_local(struct abmail *m, int n, const char **l, struct pz a[])
{
  int rc = 0;
  char *p;

  if (n > 0) {
    if ((p=strchr(m->line, ':')))
      *p = ';';
  }
  return rc;
}

#define content_type m->info[0]
#define charset m->info[1]
#define content_transfer_encoding m->info[2]
int amp_mail(struct abmail *m, int n, const char **l, struct pz a[])
{
  int rc = 0;

  if (n > 1) {
    int i;

    strlcpy(content_type, l[1], sizeof(content_type));
    m->ct = CT_unknown;
    for (i=1; i<sizeof(content_types)/sizeof(char *); i++)
      if (strcasecmp(content_type, content_types[i]) == 0) {
	m->ct = i;
	break;
      }
    if (n > 2) {
      strlcpy(charset, l[2], sizeof(charset));
      m->cs = CC_none;
      for (i=1; i<sizeof(charcodes)/sizeof(char *); i++)
	if (strcasecmp(charset, charcodes[i]) == 0) {
	  m->cs = i;
	  break;
	}
      if (!m->cs)
	if (strcasecmp(charset, "euc") == 0)
	  m->cs = CC_euc_jp;
    }
  }
  else if (!n)
    m->ct = CT_text_plain;
  return rc;
}

int amp_mailenc(struct abmail *m, int n, const char **l, struct pz a[])
{
  int rc = 0;

  if (n > 1) {
    int k;

    strlcpy(content_transfer_encoding, l[1], sizeof(content_transfer_encoding));
    m->cte = EC_none;
    for (k=1; k<sizeof(encodings)/sizeof(char *); k++)
      if (strcasecmp(content_transfer_encoding, encodings[k]) == 0) {
	m->cte = k;
	break;
      }
  }
  return rc;
}
#undef content_type
#undef charset
#undef content_transfer_encoding

int amp_boundary(struct abmail *m, int n, const char **l, struct pz a[])
{
  int rc = 0;

  if (n > 1) {
    if (m->lom < sizeof(m->boundary)/sizeof(m->boundary[0])) {
      m->lom++;
      strlcpy(m->boundary[m->lom-1], "--", sizeof(m->boundary[m->lom-1]));
      strlcat(m->boundary[m->lom-1], l[1], sizeof(m->boundary[m->lom-1]));
    }
  }
  return rc;
}

int amp_bomail(struct abmail *m, int n, const char **l, struct pz a[])
{
  m->hop = !0;
  return 0;
}

void res_read(int argc, char *argv[])
{
  char *myname = strdup(argv[0]), *herename = dirname(myname);
  char buf[2048], lst_pathname[MAXPATHLEN], *lst_basename[] = {
    "rfs_reliable.lst",
#if 0
    "ras_reliable.lst",
    "rss_reliable.lst",
    "rhs_unreliable.lst",
    "rns_reliable.lst",
    "rns_unreliable.lst",
#endif
  };
  enum {
    LST_rfs_reliable,
#if 0
    LST_ras_reliable,
    LST_rss_reliable,
    LST_rhs_unreliable,
    LST_rns_reliable,
    LST_rns_unreliable,
#endif
  };
  int i, j = 0, n = 1;
  FILE *fp;
  char *p;
  char *re;
  /*struct abmail *m = &me;*/

  for (i=0; i<sizeof(lst_basename)/sizeof(char *); i++) {
    /*int nv = 1;*/

    re = NULL;
    sprintf(lst_pathname, "%s/%s", herename, lst_basename[i]);
    if (!(fp=fopen(lst_pathname, "r"))) {
      char *home = getenv("HOME");
      sprintf(lst_pathname, "%s/.abmail/%s", home?home:"", lst_basename[i]);
      if (!(fp=fopen(lst_pathname, "r"))) {
	sprintf(lst_pathname, PREFIX"/etc/abmail/%s", lst_basename[i]);
	if (!(fp=fopen(lst_pathname, "r")))
	  ;
      }
    }
    switch (i) {
    /*case LST_rss_reliable: n = 2; break;*/
    default: n = 1; break;
    }
    while (fp && fgets(buf, sizeof(buf), fp)) {
      int c = 0/*, d = 0*/;

      if (buf[0] == '#' || buf[0] == '\n') continue;
      p = strtok(buf, "\t\n");
      while (p) {
	switch (c) {
	case 0:
	  if (!re) {
	    re = malloc(sizeof(char)*(strlen(p)+1));
	    if (re) re[0] = '\0';
	  }
	  else {
	    re = realloc(re, sizeof(char)*(strlen(re)+strlen("|")+strlen(p)+1));
	    if (re) strcat(re, "|");
	  }
	  if (re) strcat(re, p);
#if 0
	  switch (i) {
	  case LST_rss_reliable:
	    for (j=0; j<strlen(p); j++) {
	      if (p[j] == '(') d++;
	      if (strncmp(&p[j], "\\(", sizeof("\\(")-1) == 0) d--;
	      if (strncmp(&p[j], "(?:", sizeof("(?:")-1) == 0) d--;
	    }
	    m->rss = m->rss?realloc(m->rss, sizeof(struct rs)*(m->nrs+1)):
	      malloc(sizeof(struct rs)*(m->nrs+1));
	    if (m->rss) {
	      m->rss[m->nrs].lb = nv;
	      m->rss[m->nrs].ub = nv + d - 1;
	    }
	    nv += d;
	    break;
	  }
#endif
	  break;
#if 0
	case 1:
	  switch (i) {
	  case LST_rss_reliable:
	    if (m->rss) m->rss[m->nrs].se = strdup(p);
	    m->nrs++;
	    break;
	  }
	  break;
#endif
	}
	p = strtok(NULL, "\t\n");
	if (++c == n) break;
      }
    }
    if (fp) fclose(fp);
    switch (i) {
    case LST_rfs_reliable: j = HR_local; break;
#if 0
    case LST_ras_reliable: j = HR_auth; break;
    case LST_rss_reliable: j = HR_subst; headres[j].nv = nv; break;
    case LST_rhs_unreliable: j = HR_abhelo; break;
    case LST_rns_reliable: j = HR_nabhost; break;
    case LST_rns_unreliable: j = HR_abhost; break;
#endif
    }
    if ((p=malloc(sizeof(char)*(strlen(headres[j].re)-2+strlen(re?re:"")+1)))) {
      sprintf(p, headres[j].re, re?re:"");
      headres[j].re = p;
    }
    switch (i) {
    case LST_rfs_reliable: j = HR_bolocal; break;
#if 0
    case LST_ras_reliable: j = HR_boauth; break;
    case LST_rss_reliable: j = HR_bosubst; headres[j].nv = nv; break;
    case LST_rhs_unreliable: j = HR_boabhelo; break;
    case LST_rns_reliable: j = HR_bonabhost; break;
    case LST_rns_unreliable: j = HR_boabhost; break;
#endif
    }
    if ((p=malloc(sizeof(char)*(strlen(headres[j].re)-2+strlen(re?re:"")+1)))) {
      sprintf(p, headres[j].re, re?re:"");
      headres[j].re = p;
    }
    if (re) free(re);
  }
  if (myname) free(myname);
}

int abfprint(struct abmail *m, FILE *fp, char *buf)
{
#define limit_long_body(b) do {\
  char *r;\
  if (m->op.lb && !(m->op.lb - m->op.nb > 0)) {\
    rc = !0;\
    *b = '\0';\
    nb = 0;\
  }\
  else if (m->op.lb && (r = strn_wc((b), charcodes[m->ce], m->op.lb - m->op.nb, &nb))) {\
    rc = !0;\
    *r = '\0';\
  }\
} while (0)
#define limit_long_line(b) do {\
  char *r;\
  if (m->op.ll && !(m->op.ll - m->op.nl > 0)) {\
    rc = !0;\
    *b = '\0';\
    nl = 0;\
  }\
  else if (m->op.ll && (r = strnchrs((b), "\n", m->op.ll - m->op.nl, &nl))) {\
    rc = !0;\
    *(r+1) = '\0';\
  }\
} while (0)
  int rc = 0;

  if (m->op.lb && m->op.nb + 1 > m->op.lb)
    rc = !0;
  if (m->op.ll && m->op.nl + 1 > m->op.ll)
    rc = !0;
  if (!m->hda && !m->iss && !m->ibl) {
    if (rc == 0) {
      char *p;
      long nb = 0, nl = 0;

      limit_long_body(buf);
      limit_long_line(buf);
      /*fprintf(fp, "%s", buf);*/
      fputs_cc(buf,/* nl,*/ fp, charcodes[m->ceo], charcodes[m->ce]);
      if (m->op.lb) m->op.nb += nb;
      if (m->op.ll) m->op.nl += nl;
      if (buf[0] != '\0' && (p = strrchr(buf, '\0')))
	m->op.lc = *(p - 1);
    }
  }
  else if (m->iss) {
    char *p, *q;

    q = buf;
    while ((p = strnspace(q)) && rc == 0) {
      char c = '\0';
      long nb = 0;

      if ((q = strspace(p))) {
	c = *q;
	*q = '\0';
      }
      limit_long_body(p);
      if (m->hda && m->op.lc == '\n' && *p == m->op.dot) {
	fprintf(fp, "%c", m->op.dot);
	if (m->op.lb) m->op.nb++;
      }
      /*fprintf(fp, "%s", p);*/
      fputs_cc(p,/* nb,*/ fp, charcodes[m->ceo], charcodes[m->ce]);
      if (m->op.lb) m->op.nb += nb;
      m->op.lc = '\0';
      if (q) {
	if (c != '\0') {
	  fprintf(fp, " ");
	  if (m->op.lb) m->op.nb++;
	  *q = c;
	}
	q++;
      }
    }
  }
  else if (!m->ibl) {
    char *p, *q;
    long nb = 0;

    q = buf;
    while ((p = strchr(q, '\n')) && rc == 0) {
      char c = *p;

      *p = '\0';
      limit_long_body(q);
      if (m->hda && m->op.lc == '\n' && *q == m->op.dot) {
	fprintf(fp, "%c", m->op.dot);
	if (m->op.lb) m->op.nb++;
      }
      /*fprintf(fp, "%s\n", q);*/
      fputs_cc(q,/* nb,*/ fp, charcodes[m->ceo], charcodes[m->ce]);
      fputs("\n", fp);
      if (m->op.lb) m->op.nb += nb + 1;
      if (m->op.ll) {
	m->op.nl++;
	if (m->op.nl + 1 > m->op.ll)
	  rc = !0;
      }
      m->op.lc = '\n';
      *p = c;
      q = p + 1;
    }
    if (*q != '\0' && rc == 0) {
      limit_long_body(q);
      if (m->hda && m->op.lc == '\n' && *q == m->op.dot) {
	fprintf(fp, "%c", m->op.dot);
	if (m->op.lb) m->op.nb++;
      }
      /*fprintf(fp, "%s", q);*/
      fputs_cc(q,/* nb,*/ fp, charcodes[m->ceo], charcodes[m->ce]);
      if (m->op.lb) m->op.nb += nb;
      m->op.lc = *((p = strrchr(q, '\0')) - 1);
    }
  }
  else {
    char *p, *q;
    long nb = 0;

    q = buf;
    while ((p = /*strchr(q, '\n')*/strchrs(q, "\n\r")) && rc == 0) {
      char c = *p;

      *p = '\0';
      if (*q != '\0' && /*!arespace(q)*/!arespace_wc(q, charcodes[m->ce])) {
	if (m->op.lb) m->op.nb += m->op.lz;
	limit_long_body(q);
	if (m->hda && m->op.lc == '\n' && *q == m->op.dot) {
	  fprintf(fp, "%c", m->op.dot);
	  if (m->op.lb) m->op.nb++;
	}
	/*fprintf(fp, "%s%s\n", m->op.ls, q);*/
	fputs_cc(m->op.ls,/* m->op.lz, */ fp, charcodes[m->ceo], charcodes[m->ce]);
	fputs_cc(q,/* nb, */ fp, charcodes[m->ceo], charcodes[m->ce]);
	fputs("\n", fp);
	if (m->op.lb) m->op.nb += nb + 1;
	if (m->op.ll) {
	  m->op.nl++;
	  if (m->op.nl + 1 > m->op.ll)
	    rc = !0;
	}
      }
      m->op.lc = '\n';
      *p = c;
      q = p + 1;
    }
    if (*q != '\0' && rc == 0) {
      if (/*!arespace(q)*/!arespace_wc(q, charcodes[m->ce])) {
	limit_long_body(q);
	if (m->hda && m->op.lc == '\n' && *q == m->op.dot) {
	  fprintf(fp, "%c", m->op.dot);
	  if (m->op.lb) m->op.nb++;
	}
	/*fprintf(fp, "%s", q);*/
	fputs_cc(q,/* nb,*/ fp, charcodes[m->ceo], charcodes[m->ce]);
	m->op.ls[0] = '\0';
	m->op.lz = 0;
      }
      else {
	strlcpy(m->op.ls, q, sizeof(m->op.ls));
	if (m->op.lb) strn_wc(m->op.ls, charcodes[m->ce], 0, &m->op.lz);
      }
      if (m->op.ls[0] == '\0')
	m->op.lc = *((p = strrchr(q, '\0')) - 1);
      else
	m->op.lc = '\n';
    }
  }
  return rc;
}

int abbfprint(struct abmail *m, FILE *fp)
{
  int rc = 0;

  if (m->ct == CT_text_plain) {
    char buf[sizeof(m->buf)] = "";

    if (m->cte != EC_none) {
      size_t l;

      if (m->cte == EC_q && (l=strlen(m->buf))>1 && m->buf[l-2] == '=') {
	if (m->ctebuf) {
	  if ((m->ctebuf = realloc(m->ctebuf, strlen(m->ctebuf)+l+1)))
	    strcat(m->ctebuf, m->buf);
	}
	else {
	  if ((m->ctebuf = malloc(l+1)))
	    strcpy(m->ctebuf, m->buf);
	}
	if (m->ctebuf)
	  m->ctebuf[strlen(m->ctebuf)-2] = '\0';
      }
      else {
	char *src = m->buf, b[sizeof(m->buf)], *p;

	m->boe = !0;
	if (m->ctebuf) {
	  if ((m->ctebuf = realloc(m->ctebuf, strlen(m->ctebuf)+strlen(m->buf)+1)))
	    strcat(m->ctebuf, m->buf);
	  src = m->ctebuf;
	}
	if (m->csbuf) {
	  size_t s = strlen(m->csbuf);

	  strlcpy(b, m->csbuf, sizeof(b));
	  strldec_ec(b+s, src, sizeof(b)-s, strlen(src)+1, m->cte);
	}
	else
	  strldec_ec(b, src, sizeof(b), strlen(src)+1, m->cte);
	{
	  if (!(p = strrchr(b, '\n'))) {
	    if ((m->csbuf = m->csbuf ? realloc(m->csbuf, strlen(b)+1) : malloc(strlen(b)+1)))
	      strcpy(m->csbuf, b);
	  }
	  else {
	    if (*(p+1) == '\0') {
	      if (m->csbuf)
		free(m->csbuf);
	      m->csbuf = NULL;
	    }
	    else {
	      if ((m->csbuf = m->csbuf ? realloc(m->csbuf, strlen(p+1)+1) : malloc(strlen(p+1)+1)))
		strcpy(m->csbuf, p+1);
	    }
	    *(p+1) = '\0';
	    if (m->cs != CC_none && strldec_cc(buf, b, sizeof(buf), charcodes[m->ce], charcodes[m->cs]));
	    else if (strldec_wc(buf, b, sizeof(buf), charcodes[m->ce], charcodes[m->cs]));
	    else strldec_cc(buf, b, sizeof(buf), charcodes[m->ce], charcodes[CC_iso_8859_15]);
	    rc = abfprint(m, fp, buf);
	  }
	}
	if (m->ctebuf) {
	  free(m->ctebuf);
	  m->ctebuf = NULL;
	}	  
      }
    }
    else {
      if (m->cs != CC_none && strldec_cc(buf, m->buf, sizeof(buf), charcodes[m->ce], charcodes[m->cs]));
      else if (strldec_wc(buf, m->buf, sizeof(buf), charcodes[m->ce], charcodes[m->cs]));
      else strldec_cc(buf, m->buf, sizeof(buf), charcodes[m->ce], charcodes[CC_iso_8859_15]);
      rc = abfprint(m, fp, buf);
    }
  }
  return rc;
}

int abfflush(struct abmail *m, FILE *fp)
{
  int rc = 0;

  if (m->ct == CT_text_plain) {
    char buf[sizeof(m->buf)] = "";

    if (m->csbuf) {
      char *b = m->csbuf;

      if (m->cs != CC_none && strldec_cc(buf, b, sizeof(buf), charcodes[m->ce], charcodes[m->cs]));
      else if (strldec_wc(buf, b, sizeof(buf), charcodes[m->ce], charcodes[m->cs]));
      else strldec_cc(buf, b, sizeof(buf), charcodes[m->ce], charcodes[CC_iso_8859_15]);
      rc = abfprint(m, fp, buf);
      free(m->csbuf);
      m->csbuf = NULL;
    }
    if (m->op.lc != '\n')
      fprintf(fp, "\n");
    m->ct = CT_none;
    m->cte = EC_none;
    m->boe = 0;
    m->op.lc = '\0';
    m->op.ls[0] = '\0';
  }
  return rc;
}

int abhfprint(struct abmail *m, FILE *fp)
{
  struct headout {
    char *head, *item;
    long mask;
  } headouts[] = {
    { "File:",			m->filename,		OE_File			},
    { "Envelope-From",		m->from[0],		OE_Envelope_From	},
    { "Return-Path:",		m->retp[0],		OE_Return_Path		},
    { items[HI_Received],	m->item[HI_Received],	OE_Received		},
    { items[HI_Subject],	m->item[HI_Subject],	OE_Subject		},
    { items[HI_From],		m->item[HI_From],	OE_From			},
    { items[HI_To],		m->item[HI_To],		OE_To			},
    { items[HI_Message_Id],	m->item[HI_Message_Id],	OE_Message_Id		},
  };
  int i;

  if (m->op.lh) {
    long n;
    char *r;

    for (i=0; i<sizeof(m->item)/sizeof(m->item[0]); i++)
      if ((r = strn_wc(m->item[i], charcodes[m->ce], m->op.lh, &n)))
	*r = '\0';
  }
  for (i=0; i<sizeof(headouts)/sizeof(struct headout); i++)
    if (m->op.em & headouts[i].mask && headouts[i].item[0] != '\0') {
      fputs(headouts[i].head, fp);
      fputs(" ", fp);
      fputs_cc(headouts[i].item, fp, charcodes[m->ceo], charcodes[m->ce]);
      fputs("\n", fp);
      headouts[i].item[0] = '\0';
    }
  if (m->op.em & OE_eoh) fputs("\n", fp);
  m->op.lc = '\n';
  return 0;
}

struct matching {
  pcre *re;
  const char *errmsg;
  int erroffset;
  int *ovec;
  int rc;
};
struct process_headre {
  struct headre *h;
  int c, flag;
  struct matching m;
} process_headres[sizeof(headres)/sizeof(struct headre)];

int abfscan(struct abmail *m, struct process_headre *process_headres, FILE *fp)
{
  int rc = 0, eoh = 0, bob = 0, i;

  strlcpy(m->line, "", sizeof(m->line));
  while (fgets(m->buf, sizeof(m->buf), fp) != NULL && rc == 0/* && (!eoh || m->lom)*/) {
    if (!eoh) {
      size_t l;

      if ((l=strlen(m->buf)) == 0 || m->buf[l-1] != '\n') {
	;
	break;
      }
      if (m->buf[0] == '\n')
	eoh = !0;
      else if (isspace((int)m->buf[0])) {
	strlcat(m->line, m->buf, sizeof(m->line));
	continue;
      }
    }
    else {
      if (!m->body) break;
      if (m->lom && m->boundary[m->lom-1][0] != '\0') {
	if (strncmp(m->buf, m->boundary[m->lom-1], strlen(m->boundary[m->lom-1])) != 0) {
	  if (eoh)
	    rc = abbfprint(m, stdout);
	  continue;
	}
	else {
	  abfflush(m, stdout);
	  if (strncmp(&m->buf[strlen(m->boundary[m->lom-1])], "--", 2) != 0) {
	    eoh = bob = 0;
	    if (m->loh == 0)
	      m->loh = 1;
	    continue;
	  } else
	    if (!--m->lom)
	      break;
	}
      }
      else
	rc = abbfprint(m, stdout);
      strlcpy(m->line, m->buf, sizeof(m->line));
    }
    for (i=0; i<sizeof(headres)/sizeof(struct headre) && rc == 0 && (!bob || m->loh > 0); i++) {
      struct process_headre *p = &process_headres[i];

      if (p->h->sw - m->loh != 1) continue;
      if (!p->m.re) continue;
      if (strncasecmp(p->h->header, m->line, strlen(p->h->header)) == 0) {
	if (p->h->n < 0 || (p->h->n == p->c)) {
	  p->m.rc = pcre_exec(p->m.re, NULL,
			      m->line, sizeof(m->line), strlen(p->h->header),
			      0, p->m.ovec, p->h->nv*3);
	  if (p->m.rc >= 0) {
	    if (!p->h->proc) {
	      if (!p->h->neg)
		;
	    }
	    else {
	      const char **l;
	      struct pz a[p->h->nv];

	      if (!pcre_get_substring_list(m->line, p->m.ovec, p->m.rc, &l)) {
		pcre_set_substring_pz_array(m->line, p->m.ovec, p->m.rc, a);
		rc = p->h->proc(m, p->m.rc, l, a);
		pcre_free_substring_list(l);
	      }
	    }
	    if (p->h->neg)
	      p->flag = 0;
	  }
	}
	p->c++;
      }
    }
    if (eoh && !bob)
      switch (m->loh) {
      case 2: fputs("_\n", stdout);
      case 0:
	abhfprint(m, stdout);
	break;
      }
    if (!m->lom && rc != 0) break;
    if (eoh && !bob) {
      for (i=0; i<sizeof(headres)/sizeof(struct headre) && rc == 0; i++) {
	struct process_headre *p = &process_headres[i];

	p->c = 0;
	if (p->h->sw - m->loh != 1) continue;
	if (!p->m.re) continue;
	if (!p->h->neg) continue;
	if (p->flag) {
	  if (!p->h->proc)
	    ;
	  else
	    rc = p->h->proc(m, 0, NULL, NULL);
	}
      }
      bob = !0;
      if (m->loh == 1 && m->hop) {
	eoh = bob = 0;
	m->loh = 2;
      }
      else if (m->loh == 2) {
	/*m->lom = */m->hop = 0;
	m->loh = 1;
      }
    }
    strlcpy(m->line, m->buf, sizeof(m->line));
  }
  abfflush(m, stdout);
  if (m->hop) ;
  ;
  for (i=0; i<sizeof(headres)/sizeof(struct headre); i++) {
    struct process_headre *p = &process_headres[i];

    p->c = 0;
    p->flag = !0;
  }
  return rc;
}

char *mailpath(const char *pathname, size_t *sz, int *mailpathtype)
{
  int i, is_maildir = 0;
  char *rv, *maildirs[] = { "cur", "tmp", "new" }; /* last should be "new" */
  size_t l = pathname ? strlen(pathname) : 0,
    z = (l ? l+1 : 0) + max(max(sizeof("cur"), sizeof("tmp")), sizeof("new"));
  struct stat sb;

  if (sz)
    *sz = l+1;
  if ((rv = malloc(z))) {
    for (i=0; i<sizeof(maildirs)/sizeof(char *); i++) {
      strcpy(rv, "");
      if (l) {
	strcat(rv, pathname);
	strcat(rv, "/");
      }
      strcat(rv, maildirs[i]);
      if (!(is_maildir=(stat(rv, &sb) == 0 && S_ISDIR(sb.st_mode))))
	break;
    }
    if (is_maildir)
      *sz = z;
    else {
      free(rv);
      rv = pathname ? strdup(pathname) : NULL;
    }      
  }
  if (mailpathtype)
    *mailpathtype = is_maildir;
  return rv;
}

int main(int argc, char *argv[])
{
  int i, a;
  int exist = 0;
  FILE *fp = NULL;
  long c = 0;

  for (a=1; a<argc; a++)
    if (strcmp(argv[a], "--") == 0)
      break;
    else if (strcmp(argv[a], "-E") == 0)
      exist = !0;
    else if (strcmp(argv[a], "-b-") == 0)
      me.body = 0;
    else if (strcmp(argv[a], "-D-") == 0)
      me.hda = 0;
    else if (strcmp(argv[a], "-S") == 0)
      me.iss = !0;
    else if (strcmp(argv[a], "-B-") == 0)
      me.ibl = 0;
    else if (strcmp(argv[a], "-s") == 0)
      me.op.em = OE_File | OE_From | OE_Subject;
    else if (strcmp(argv[a], "-w") == 0 && aredigit(argv[a+1]))
      me.op.lh = atol(argv[++a]);
    else if (strcmp(argv[a], "-c") == 0 && aredigit(argv[a+1]))
      me.op.lb = atol(argv[++a]);
    else if (strcmp(argv[a], "-l") == 0 && aredigit(argv[a+1]))
      me.op.ll = atol(argv[++a]);
    else if (strcmp(argv[a], "-") == 0)
      fp = stdin;
    else
      break;
  if (exist) {
    for (; a<argc; a++) {
      struct stat sb;

      if (stat(argv[a], &sb) != 0)
	continue;
      if (S_ISDIR(sb.st_mode)) {
	DIR *dir;
	struct dirent *d;
	size_t z;
	char *mailpathname;
	int mailpathtype;

	mailpathname = mailpath(argv[a], &z, &mailpathtype);
	if (!(dir = opendir(mailpathname))) {
	  if (mailpathname) free(mailpathname);
	  continue;
	}
	while ((d = readdir(dir))) { /* ! */
	  char *pathname;

	  if (!mailpathtype && !aredigit(d->d_name))
	    continue;
	  if ((pathname = malloc(z + 1 + strlen(d->d_name)))) {
	    strcpy(pathname, mailpathname);
	    strcat(pathname, "/");
	    strcat(pathname, d->d_name);
	  }
	  if (stat(pathname, &sb) == 0 && S_ISDIR(sb.st_mode)) {
	    if (pathname) free(pathname);
	    continue;
	  }
	  {
	    if (!(fp = fopen(pathname, "r"))) {
	      if (pathname) free(pathname);
	      continue;
	    }
	    c++;
	    break;
	  }
	  if (pathname) free(pathname);
	}
	closedir(dir);
	if (mailpathname) free(mailpathname);
	if (c) break;
	continue;
      }
      {
	if (!(fp = fopen(argv[a], "r")))
	  continue;
	fclose(fp);
	c++;
	break;
      }
    }
    return c ? 0 : 1;
  }
  res_read(argc, argv);
  for (i=0; i<sizeof(headres)/sizeof(struct headre); i++) {
    struct process_headre *p = &process_headres[i];

    p->h = &headres[i];
    p->c = 0;
    p->flag = !0;
    if (p->h->sw == 0) continue;
    p->m.re = pcre_compile(p->h->re,
			   PCRE_ANCHORED|PCRE_CASELESS,
			   &p->m.errmsg, &p->m.erroffset, NULL);
    if (!p->m.re)
      fprintf(stderr, "pcre compilation failed at offset %d: %s\n%s\n",
	      p->m.erroffset, p->m.errmsg, p->h->re);
    p->h->nv = max(10, p->h->nv);
    if (!(p->m.ovec = malloc(sizeof(int)*p->h->nv*3)))
      p->h->nv = 0;
  }
  if (fp) {
    struct abmail m = me;

    strldec_cc(m.filename, "-", sizeof(m.filename), charcodes[m.ce], "");
    abfscan(&m, process_headres, fp);
    fprintf(stdout, "%c\n", m.op.dot);
    c++;
  }
  {
    for (; a<argc; a++) {
      struct stat sb;

      if (stat(argv[a], &sb) != 0)
	continue;
      if (S_ISDIR(sb.st_mode)) {
	DIR *dir;
	struct dirent *d;
	size_t z;
	char *mailpathname;
	int mailpathtype;

	mailpathname = mailpath(argv[a], &z, &mailpathtype);
	if (!(dir = opendir(mailpathname))) {
	  if (mailpathname) free(mailpathname);
	  continue;
	}
	while ((d = readdir(dir))) { /* ! */
	  char *pathname;

	  if (!mailpathtype && !aredigit(d->d_name))
	    continue;
	  if ((pathname = malloc(z + 1 + strlen(d->d_name)))) {
	    strcpy(pathname, mailpathname);
	    strcat(pathname, "/");
	    strcat(pathname, d->d_name);
	  }
	  if (stat(pathname, &sb) == 0 && S_ISDIR(sb.st_mode)) {
	    if (pathname) free(pathname);
	    continue;
	  }
	  {
	    struct abmail m = me;

	    if (!(fp = fopen(pathname, "r"))) {
	      if (pathname) free(pathname);
	      continue;
	    }
	    strldec_cc(m.filename, pathname, sizeof(m.filename), charcodes[m.ce], "");
	    abfscan(&m, process_headres, fp);
	    fprintf(stdout, "%c\n", m.op.dot);
	    fclose(fp);
	    c++;
	  }
	  if (pathname) free(pathname);
	}
	closedir(dir);
	if (mailpathname) free(mailpathname);
	continue;
      }
      {
	struct abmail m = me;

	if (!(fp = fopen(argv[a], "r")))
	  continue;
	strldec_cc(m.filename, argv[a], sizeof(m.filename), charcodes[m.ce], "");
	abfscan(&m, process_headres, fp);
	fprintf(stdout, "%c\n", m.op.dot);
	fclose(fp);
	c++;
      }
    }
  }
  if (c > 1) fprintf(stdout, "FileCount: %ld\n", c);
  return 0;
}
