/*
 * ararray-bench.cc - benchmark for arithmetic-array class template
 *
 * Copyright (C) 2009 Taiji Yamada <taiji@aihara.co.jp>
 */
#include <inttypes.h>	/* uint32_t, etc */
#include <assert.h>
#include <iostream>
#include <cstdlib>
#include <ararray.hpp>	/* arithmetic-array class template */
#include <gmpxx.h>	/* GNU MP for C++ */
inline bool operator!(const mpz_class a)
{
  return a == 0;
}
#include <sys/time.h>	/* gettimeofday, etc */
#include <sys/resource.h>	/* getrusage, etc */
using namespace std;

class StopWatch {
public:
  void start()
  {
    gettimeofday(&tv, NULL);
    ctime[0] = tv.tv_sec + 1E-6*tv.tv_usec;
    getrusage(RUSAGE_SELF, &ru);
    utime[0] = ru.ru_utime.tv_sec + 1E-6*ru.ru_utime.tv_usec;
  }
  void stop()
  {
    gettimeofday(&tv, NULL);
    ctime[1] = tv.tv_sec + 1E-6*tv.tv_usec;
    getrusage(RUSAGE_SELF, &ru);
    utime[1] = ru.ru_utime.tv_sec + 1E-6*ru.ru_utime.tv_usec;
  }
  const double get_ctime() const
  {
    return ctime[1] - ctime[0];
  }
  const double get_utime() const
  {
    return utime[1] - utime[0];
  }
  void watch() const
  {
    std::cout << "ctime: " << get_ctime() << "\n";
    std::cout << "utime: " << get_utime() << "\n";
  }
private:
  struct timeval tv;
  double ctime[2];
  struct rusage ru;
  double utime[2];
};

#if 1
#define RAND random()
#define SRAND srandom()
#if defined(sun)
#define SRANDDEV srandom(getpid())
#else
#define SRANDDEV srandomdev()
#endif
#else
#define RAND rand()
#define SRAND srand()
#if defined(sun)
#define SRANDDEV srand(getpid())
#else
#define SRANDDEV sranddev()
#endif
#endif

#if defined(__LP64__)
#define get_val get_ull
#define longes_t uint64_t
#define A_elem_t uint64_t
#define B_elem_t uint32_t
#define C_elem_t uint16_t
#else
#define get_val get_ul
#define longes_t uint32_t
#define A_elem_t uint32_t
#define B_elem_t uint16_t
#define C_elem_t uint8_t
#endif

namespace N {
  longes_t a, b, c, d;
  bool f;
  double bm;
}

#define A_elem_n sizeof(longes_t)/sizeof(A_elem_t)
namespace A {
  ararray<A_elem_t, A_elem_n> a, b, c, d;
  bool f;
  double bm;
}

#define B_elem_n sizeof(longes_t)/sizeof(B_elem_t)
namespace B {
  ararray<B_elem_t, B_elem_n> a, b, c, d;
  bool f;
  double bm;
}

#define C_elem_n sizeof(longes_t)/sizeof(C_elem_t)
namespace C {
  ararray<C_elem_t, C_elem_n> a, b, c, d;
  bool f;
  double bm;
}

namespace M {
  mpz_class a, b, c, d;
  bool f;
  double bm;
}

#define do_check(T) do {\
  using namespace T;\
  assert(N::a	== a.get_val());\
  /* wrong at gcc-4.0.1 -O1 -arch ppc64 *//*\
  assert(N::a++	== a++.get_val());\
  assert(N::a++	== a++.get_val());\
  assert(N::a--	== a--.get_val());\
  assert(N::a--	== a--.get_val());\
  */\
  assert((N::a++, a++, N::a	== a.get_val()));\
  assert((N::a--, a--, N::a	== a.get_val()));\
  assert(!N::a	== !a);\
  assert(~N::a	== ~a.get_val());\
  assert(++N::a	== (++a).get_val());\
  assert(--N::a	== (--a).get_val());\
  N::a = N::c; a = c; N::b = N::d; b = d;\
  assert(N::a + N::b	== (a + b).get_val());\
  assert(N::a - N::b	== (a - b).get_val());\
  assert((N::a & N::b)	== (a & b).get_val());\
  assert((N::a ^ N::b)	== (a ^ b).get_val());\
  assert((N::a | N::b)	== (a | b).get_val());\
  N::a = N::c; a = c; N::b = N::d; b = d;\
  assert((N::a << s)		== (a << s).get_val());\
  assert(((N::a << s) >> s)	== ((a << s) >> s).get_val());\
  N::a = N::c; a = c; N::b = N::d; b = d;\
  assert((N::a += N::b)	== (a += b).get_val());\
  N::a = N::c; a = c; N::b = N::d; b = d;\
  assert((N::a -= N::b)	== (a -= b).get_val());\
  N::a = N::c; a = c; N::b = N::d; b = d;\
  assert((N::a &= N::b)	== (a &= b).get_val());\
  N::a = N::c; a = c; N::b = N::d; b = d;\
  assert((N::a ^= N::b)	== (a ^= b).get_val());\
  N::a = N::c; a = c; N::b = N::d; b = d;\
  assert((N::a |= N::b)	== (a |= b).get_val());\
  N::a = N::c; a = c; N::b = N::d; b = d;\
  assert((N::a <= N::b)	== (a <= b));\
  assert((N::a < N::b)	== (a < b));\
  assert((N::a >= N::b)	== (a >= b));\
  assert((N::a > N::b)	== (a > b));\
  assert((N::a == N::b)	== (a == b));\
  assert((N::a != N::b)	== (a != b));\
  N::a = N::c; a = c; N::b = N::d; b = d;\
  assert((N::a <= N::b)	== (a <= b));\
  assert((N::a < N::b)	== (a < b));\
  assert((N::a >= N::b)	== (a >= b));\
  assert((N::a > N::b)	== (a > b));\
  assert((N::a == N::b)	== (a == b));\
  assert((N::a != N::b)	== (a != b));\
} while (0)

void check()
{
  StopWatch sw;
  const long niter = 1000000;
  const int s = 20;
  cout << "ararray.hpp check:\t"
    "ararray(T,n)s compared with native uint" << endl << "checking ";
  sw.start();
  for (long i=0; i<niter; i++) {
    N::c = N::a = RAND;
    N::d = N::b = RAND;

    A::c = A::a = N::a;
    A::d = A::b = N::b;

    B::c = B::a = N::a;
    B::d = B::b = N::b;

    C::c = C::a = N::a;
    C::d = C::b = N::b;

    do_check(A);
    do_check(B);
    do_check(C);
  }
  sw.stop();
  cout << "status:\tsuccessful" << endl;
  cout << "operation set times:\t" << niter <<
    " * (3 ararray(T,n)s and 1 native uint)" << endl;
  cout << "total time(sec):\t" << sw.get_ctime() << endl << endl;
}

#define do_loop(exp) do {\
  long i;\
  sw.start();\
  for (i=0; i<niter; i++)\
    exp;\
  sw.stop();\
  cout << niter/sw.get_ctime() << "\t";\
  bm = (k*bm + niter/sw.get_ctime())/(k+1);\
} while (0)

#define do_bench(exp,ass) do {\
  N::c = N::a = RAND;\
  N::d = N::b = RAND;\
  A::c = A::a = N::a;\
  A::d = A::b = N::b;\
  B::c = B::a = N::a;\
  B::d = B::b = N::b;\
  C::c = C::a = N::a;\
  C::d = C::b = N::b;\
  M::c = M::a = (uint32_t)N::a;\
  M::d = M::b = (uint32_t)N::b;\
  cout << #exp ":\t";\
  { using namespace N; do_loop(exp); }\
  { using namespace A; do_loop(exp); assert(ass); }\
  { using namespace B; do_loop(exp); assert(ass); }\
  { using namespace C; do_loop(exp); assert(ass); }\
  { using namespace M; do_loop(exp); }\
  cout << endl;\
  k++;\
} while (0)

void markbench()
{
  using namespace std;
  StopWatch sw;
  const long niter = 1000000;
  const int s = 20;
  int k = 0;
  {
    N::bm = A::bm = B::bm = C::bm = M::bm = 0;
    cout << "ararray.hpp bench:\t"
      "ararray(T,n)s compared with native uint and mpz_class" << endl;
    cout << "unit:\toperation times/sec" << endl;
    cout << "op.:\tuint" << 8*sizeof(longes_t) << "\t\t" <<
      "(uint" << 8*sizeof(A_elem_t)<< "," << A_elem_n << ")\t" <<
      "(uint" << 8*sizeof(B_elem_t)<< "," << B_elem_n << ")\t" <<
      "(uint" << 8*sizeof(C_elem_t)<< "," << C_elem_n << ")\t" <<
      "mpz_class\t" << endl;
    do_bench(a++,	N::a == a.get_val());
    do_bench(a--,	N::a == a.get_val());
    do_bench(f=!a,	N::f == f);
    do_bench(b=~a,	N::b == b.get_val());
    do_bench(++a,	N::a == a.get_val());
    do_bench(--a,	N::a == a.get_val());
    do_bench(b=a+b,	N::b == b.get_val());
    do_bench(b=a-b,	N::b == b.get_val());
    do_bench(b=a&b,	N::b == b.get_val());
    do_bench(b=a^b,	N::b == b.get_val());
    do_bench(b=a|b,	N::b == b.get_val());
    do_bench(b=a<<s,	N::b == b.get_val());
    do_bench(b=a>>s,	N::b == b.get_val());
    do_bench(a+=b,	N::a == a.get_val());
    do_bench(a-=b,	N::a == a.get_val());
    do_bench(a&=b,	N::a == a.get_val());
    do_bench(a^=b,	N::a == a.get_val());
    do_bench(a|=b,	N::a == a.get_val());
    do_bench(f=a<=b,	N::f == f);
    do_bench(f=a<b,	N::f == f);
    do_bench(f=a>=b,	N::f == f);
    do_bench(f=a>b,	N::f == f);
    do_bench(f=a==b,	N::f == f);
    do_bench(f=a!=b,	N::f == f);
    cout << "checking status:\tsuccessfull" << endl;
    cout << "ave.:\t" << N::bm << "\t" <<
      A::bm << "\t" <<
      B::bm << "\t" <<
      C::bm << "\t" <<
      M::bm << "\t" << endl;
  }
}

int main()
{
  SRANDDEV;
  check();
  markbench();
  return 0;
}
