/*
 * arr_ar.h - arithmetic macros for array
 *
 * Copyright (C) 2007,2009 Taiji Yamada <taiji@aihara.co.jp>
 *
 * Distributed under the Boost Software License, Version 1.0. See
 * http://www.boost.org/LICENSE_1_0.txt for more details.
 */
#define arr_init(n, a) do {\
  memset(&(a)[0], 0, sizeof((a)[0])*n);	\
} while (0)

#define arr_not(n, a, b) do {\
  int i;\
\
  for (i=0; i<n; i++)\
    (b)[i] = ~(a)[i];\
} while (0)

#define arr_and(n, a, b, c) do {\
  int i;\
\
  for (i=0; i<n; i++)\
    (c)[i] = (a)[i] & (b)[i];\
} while (0)

#define arr_or(n, a, b, c) do {\
  int i;\
\
  for (i=0; i<n; i++)\
    (c)[i] = (a)[i] | (b)[i];\
} while (0)

#define arr_xor(n, a, b, c) do {\
  int i;\
\
  for (i=0; i<n; i++)\
    (c)[i] = (a)[i] ^ (b)[i];\
} while (0)

#define arr_lsr(n, a, s, b) do {\
  int i;\
  const int d = s / (8*sizeof((a)[0])), m = s % (8*sizeof((a)[0])),\
    u = 8*sizeof((a)[0]) - m;\
\
  for (i=n-1; i>=0; i--) {\
    int k = i - d, l = k - 1;\
\
    (b)[i] = ((0<=k && k<n) ? (a)[k]>>m : 0) | ((0<=l && l<n) ? (a)[l]<<u : 0);\
  }\
} while (0)

#define arr_lsl(n, a, s, b) do {\
  int i;\
  const int d = s / (8*sizeof((a)[0])), m = s % (8*sizeof((a)[0])),\
    u = 8*sizeof((a)[0]) - m;\
\
  for (i=0; i<n; i++) {\
    int k = i + d, l = k + 1;\
\
    (b)[i] = ((0<=k && k<n) ? (a)[k]<<m : 0) | ((0<=l && l<n) ? (a)[l]>>u : 0);\
  }\
} while (0)

#if 0 /* simple version for addition and subtraction */

#define arr_add(n, a, b, c) do {\
  int i;\
\
  for (i=n-1; i>=0; i--) {\
    if (i>0) (c)[i-1] = (c)[i];\
    (c)[i] = (a)[i] + (b)[i] + (c)[i];\
    if (i>0) (c)[i-1] = (c)[i-1] ? ((c)[i]<=(a)[i] ? 1 : 0) : ((c)[i]<(a)[i] ? 1 : 0);\
  }\
} while (0)

#define arr_sub(n, a, b, c) do {\
  int i;\
\
  for (i=n-1; i>=0; i--) {\
    if (i>0) (c)[i-1] = (c)[i];\
    (c)[i] = (a)[i] - (b)[i] - (c)[i];\
    if (i>0) (c)[i-1] = (c)[i-1] ? ((c)[i]>=(a)[i] ? 1 : 0) : ((c)[i]>(a)[i] ? 1 : 0);\
  }\
} while (0)

#define arr_inc(n, a, c) do {\
  int i;\
\
  for (i=n-1; i>=0; i--) {\
    if (i>0) (c)[i-1] = (c)[i];\
    (c)[i] = (a)[i] + ((i == n-1) ? 1 : 0) + (c)[i];\
    if (i>0) (c)[i-1] = (c)[i-1] ? ((c)[i]<=(a)[i] ? 1 : 0) : ((c)[i]<(a)[i] ? 1 : 0);\
  }\
} while (0)

#define arr_dec(n, a, c) do {\
  int i;\
\
  for (i=n-1; i>=0; i--) {\
    if (i>0) (c)[i-1] = (c)[i];\
    (c)[i] = (a)[i] - ((i == n-1) ? 1 : 0) - (c)[i];\
    if (i>0) (c)[i-1] = (c)[i-1] ? ((c)[i]>=(a)[i] ? 1 : 0) : ((c)[i]>(a)[i] ? 1 : 0);\
  }\
} while (0)

#else

#define arr_add(n, a, b, c) do {\
  int i = n-1;\
\
  if (i>0) {\
    (c)[i] = (a)[i] + (b)[i];\
    (c)[i-1] = (c)[i]<(a)[i] ? 1 : 0;\
    i--;\
  }\
  for (; i>0; i--) {\
    (c)[i] = (a)[i] + (b)[i] + ((c)[i-1]=(c)[i]);\
    (c)[i-1] = (c)[i-1] ? ((c)[i]<=(a)[i] ? 1 : 0) : ((c)[i]<(a)[i] ? 1 : 0);\
  }\
  if (i>=0)\
    (c)[i] = (a)[i] + (b)[i] + (c)[i];\
} while (0)

#define arr_sub(n, a, b, c) do {\
  int i = n-1;\
\
  if (i>0) {\
    (c)[i] = (a)[i] - (b)[i];\
    (c)[i-1] = (c)[i]>(a)[i] ? 1 : 0;\
    i--;\
  }\
  for (; i>0; i--) {\
    (c)[i] = (a)[i] - (b)[i] - ((c)[i-1]=(c)[i]);\
    (c)[i-1] = (c)[i-1] ? ((c)[i]>=(a)[i] ? 1 : 0) : ((c)[i]>(a)[i] ? 1 : 0);\
  }\
  if (i>=0)\
    (c)[i] = (a)[i] - (b)[i] - (c)[i];\
} while (0)

#define arr_inc(n, a, c) do {\
  int i = n-1;\
\
  if (n == 1) {\
    (c)[i] = (a)[i] + 1;\
    break;\
  }\
  if (i>0) {\
    (c)[i] = (a)[i] + 1;\
    (c)[i-1] = (c)[i]<(a)[i] ? 1 : 0;\
    i--;\
  }\
  for (; i>0; i--) {\
    (c)[i] = (a)[i] + ((c)[i-1]=(c)[i]);\
    (c)[i-1] = (c)[i-1] ? ((c)[i]<=(a)[i] ? 1 : 0) : ((c)[i]<(a)[i] ? 1 : 0);\
  }\
  if (i>=0)\
    (c)[i] = (a)[i] + (c)[i];\
} while (0)

#define arr_dec(n, a, c) do {\
  int i = n-1;\
\
  if (n == 1) {\
    (c)[i] = (a)[i] - 1;\
    break;\
  }\
  if (i>0) {\
    (c)[i] = (a)[i] - 1;\
    (c)[i-1] = (c)[i]>(a)[i] ? 1 : 0;\
    i--;\
  }\
  for (; i>0; i--) {\
    (c)[i] = (a)[i] - ((c)[i-1]=(c)[i]);\
    (c)[i-1] = (c)[i-1] ? ((c)[i]>=(a)[i] ? 1 : 0) : ((c)[i]>(a)[i] ? 1 : 0);\
  }\
  if (i>=0)\
    (c)[i] = (a)[i] - (c)[i];\
} while (0)

#endif

#define arr_each(n, exp) do {\
  int i;\
\
  for (i=0; i<n; i++)\
    exp;\
} while (0)
#define arr_tr(n, a, tr, b) arr_each(n, (b)[i] = tr((a)[i]))
#define arr_is(n, exp) arr_each(n, if (exp) break)
#define arr_are(n, exp) arr_each(n, if (!(exp)) break)

#define arr_each_rev(n, exp) do {\
  int i;\
\
  for (i=n-1; i>=0; i--)\
    exp;\
} while (0)
#define arr_tr_rev(n, a, tr, b) arr_each_rev(n, (b)[i] = tr((a)[i]))
#define arr_is_rev(n, exp) arr_each_rev(n, if (exp) break)
#define arr_are_rev(n, exp) arr_each_rev(n, if (!(exp)) break)

#define arr_fprint_bits(n, a, fp) do {\
  int i, j;\
\
  for (i=0; i<n; i++) {\
    for (j=0; j<8*sizeof((a)[0]); j++)\
      if ((a)[i] & 1<<(8*sizeof((a)[0])-j-1))\
	fprintf(fp, "1");\
      else\
	fprintf(fp, "0");\
    if (i%2) fprintf(fp, i!=n-1?":":"\n");\
  }\
} while (0)

/* swap for arbitrary type */
#define art_swap(a,b) do {\
  struct tmp_t { char t[sizeof(a)]; } tmp;\
  tmp = *(struct tmp_t *)&(a);\
  *(struct tmp_t *)&(a) = *(struct tmp_t *)&(b);\
  *(struct tmp_t *)&(b) = tmp;\
} while(0)

/* change endian for arbitrary type */
#define art_chend(p) do {\
  size_t i;\
  for (i=0; i<sizeof(*(p))/2; i++)\
    art_swap(((char *)p)[i], ((char *)p)[sizeof(*(p))-i-1] );\
} while (0)
